#include "ogreframework.h"

#include "../OgreSDK/FileSystemLayerImpl.h"

#ifndef OGRE_PLUGINS_FILE
#define OGRE_PLUGINS_FILE "plugins.cfg"
#endif

#ifndef OGRE_CONFIG_FILE
#define OGRE_CONFIG_FILE "ogre.cfg"
#endif

#ifndef OGRE_LOG_FILE
#define OGRE_LOG_FILE "ogre.log"
#endif

#ifndef OGRE_RESOURCES_FILE
#define OGRE_RESOURCES_FILE "resources.cfg"
#endif

using namespace Ogre;

template<> OgreFramework* Ogre::Singleton<OgreFramework>::ms_Singleton = 0;

OgreFramework::OgreFramework()
{
    m_root = NULL;
    m_renderWindow = NULL;
    m_viewport = NULL;
    m_timer = NULL;

    m_inputManager = NULL;
    m_keyboard = NULL;
    m_mouse = NULL;
}

OgreFramework::~OgreFramework()
{
    if (m_inputManager)
        OIS::InputManager::destroyInputSystem(m_inputManager);

    if (m_root)
        delete m_root;
    
#ifdef GAME_STATIC_BUILD
    delete m_glPlugin;
    delete m_cgPlugin;
    delete m_particleFXPlugin;
    delete m_soundPlugin;
#endif
}

bool OgreFramework::initOgre(Ogre::String windowTitle, OIS::KeyListener *keyListener, OIS::MouseListener *mouseListener)
{
    // Get platform independent paths to user's files
    OgreBites::FileSystemLayerImpl fs("velocityoverflow");
    Ogre::String configFile = fs.getWritablePath(OGRE_CONFIG_FILE);
    Ogre::String logFile = fs.getWritablePath(OGRE_LOG_FILE);

    m_root = new Ogre::Root(OGRE_PLUGINS_FILE, configFile, logFile);

    // Install plugins manually (if building statically)
    installPlugins();
    
    if (!m_root->showConfigDialog()) {
        return false;
    }

    m_renderWindow = m_root->initialise(true, windowTitle);

    // Note: Z-order (2) of each viewport must be different from the others
    m_viewport = m_renderWindow->addViewport(NULL, 2);
    m_viewport->setBackgroundColour(Ogre::ColourValue(0.0, 0.0, 0.0));
    m_viewport->setCamera(NULL);

    unsigned long hWnd = 0;
    OIS::ParamList paramList;
    m_renderWindow->getCustomAttribute("WINDOW", &hWnd);

    paramList.insert(OIS::ParamList::value_type("WINDOW", Ogre::StringConverter::toString(hWnd)));

    // Non-exclusive input; just for debugging
#if (defined OIS_WIN32_PLATFORM) && (defined GAME_DEBUG)
paramList.insert(std::make_pair(std::string("w32_mouse"), std::string("DISCL_FOREGROUND" )));
//paramList.insert(std::make_pair(std::string("w32_mouse"), std::string("DISCL_NONEXCLUSIVE")));
paramList.insert(std::make_pair(std::string("w32_mouse"), std::string("DISCL_EXCLUSIVE")));
paramList.insert(std::make_pair(std::string("w32_keyboard"), std::string("DISCL_FOREGROUND")));
paramList.insert(std::make_pair(std::string("w32_keyboard"), std::string("DISCL_NONEXCLUSIVE")));
#elif (defined OIS_LINUX_PLATFORM) && (defined GAME_DEBUG)
paramList.insert(std::make_pair(std::string("x11_mouse_grab"), std::string("false")));
paramList.insert(std::make_pair(std::string("x11_mouse_hide"), std::string("false")));
paramList.insert(std::make_pair(std::string("x11_keyboard_grab"), std::string("false")));
paramList.insert(std::make_pair(std::string("XAutoRepeatOn"), std::string("true")));
#endif

    m_inputManager = OIS::InputManager::createInputSystem(paramList);

    m_keyboard = static_cast<OIS::Keyboard*>(m_inputManager->createInputObject(OIS::OISKeyboard, true));
    m_mouse = static_cast<OIS::Mouse*>(m_inputManager->createInputObject(OIS::OISMouse, true));

    m_mouse->getMouseState().height = m_renderWindow->getHeight();
    m_mouse->getMouseState().width  = m_renderWindow->getWidth();

    if (keyListener == NULL)
        m_keyboard->setEventCallback(this);
    else
        m_keyboard->setEventCallback(keyListener);

    if (mouseListener = NULL)
        m_mouse->setEventCallback(this);
    else
        m_mouse->setEventCallback(mouseListener);

    loadResources();

    m_timer = new Ogre::Timer;
    m_timer->reset();

    m_renderWindow->setActive(true);

    return true;
}

void OgreFramework::installPlugins()
{
#ifdef GAME_STATIC_BUILD
    m_glPlugin = new GLPlugin();
    m_cgPlugin = new CgPlugin();
    m_particleFXPlugin = new ParticleFXPlugin();
    m_soundPlugin = new OgreOggSound::OgreOggSoundPlugin();
    
    m_root->installPlugin(m_glPlugin);
    m_root->installPlugin(m_cgPlugin);
    m_root->installPlugin(m_particleFXPlugin);
    m_root->installPlugin(m_soundPlugin);
#endif
}

void OgreFramework::loadResources()
{
    // Load resource file
    Ogre::ConfigFile cf;
    cf.load(OGRE_RESOURCES_FILE);

    Ogre::ConfigFile::SectionIterator sectionIter = cf.getSectionIterator();
    Ogre::String sectionName, typeName, dataName;

    // Iterate over sections
    while (sectionIter.hasMoreElements()) {
        sectionName = sectionIter.peekNextKey();

        Ogre::ConfigFile::SettingsMultiMap *settings = sectionIter.getNext();
        Ogre::ConfigFile::SettingsMultiMap::iterator i;

        // Iterate over each setting
        for (i = settings->begin(); i != settings->end(); ++i) {
            typeName = i->first;
            dataName = i->second;

            // Add data to resource index
            Ogre::ResourceGroupManager::getSingleton().
                    addResourceLocation(dataName, typeName, sectionName);
        }
    }

    Ogre::ResourceGroupManager::getSingleton().initialiseAllResourceGroups();
}
