/**
 *  @author Marian Ondrejicek (xondre02)
 *  @date   May 2012
 *
 *  Manages all in-game cameras.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "cameramanager.h"

CameraManager::CameraManager(Ogre::SceneManager *manager)
{
    m_sceneManager = manager;
}

CameraManager::~CameraManager()
{
    // Destroy all cameras
    for (unsigned int i = 0; i < m_cameras.size(); i++) {
        GameCamera *gameCam = m_cameras[i].camera;
        m_sceneManager->destroyCamera(gameCam->getCamera());
        delete gameCam;
    }
}

GameCamera *CameraManager::createCamera(int type, Ogre::Real aspect, int player)
{
    // Create Ogre::Camera with unique name
    static int counter = 0;
    std::string camName = "GameCamera" + Ogre::StringConverter::toString(counter++);
    Ogre::Camera *cam = m_sceneManager->createCamera(camName);
    cam->setNearClipDistance(5);
    cam->setAspectRatio(aspect);

    // Choose GameCamera subclass according to type
    GameCamera *gameCam;
    if (type == CAMERA_DYNAMIC)
        gameCam = new DynamicCamera;
    else if (type == CAMERA_SIDE)
        gameCam = new SideCamera;
    else if (type == CAMERA_STATIC)
        gameCam = new StaticCamera;
    else
        gameCam = new GameCamera;

    // Wrap the camera around GameCamera object
    gameCam->setCamera(cam);

    // Save it to PlayerCamera structure with other information
    PlayerCamera playerCamera;
    playerCamera.camera = gameCam;
    playerCamera.type = type;
    playerCamera.player = player;

    // Add it to the list
    m_cameras.push_back(playerCamera);

    // Return pointer to the created GameCamera object
    return gameCam;
}

GameCamera *CameraManager::getCamera(int type, int player)
{
    // Try to find the camera
    for (unsigned int i = 0; i < m_cameras.size(); i++) {
        if (m_cameras[i].type == type && m_cameras[i].player == player) {
            return m_cameras[i].camera;
        }
    }

    return NULL;
}

GameCamera *CameraManager::getCamera(int type, int player, int n)
{
    // Try to find the n-th camera of given type
    int counter = 0;

    for (unsigned int i = 0; i < m_cameras.size(); i++) {
        if (m_cameras[i].type == type && m_cameras[i].player == player) {
            if (counter++ == n)
                return m_cameras[i].camera;
        }
    }

    return NULL;
}

void CameraManager::update(double timeSinceLastFrame, const Ogre::SceneNode *target, int player)
{
    // Pass the information to all cameras of the player
    for (unsigned int i = 0; i < m_cameras.size(); i++) {
        if (m_cameras[i].player == player) {
            m_cameras[i].camera->update(timeSinceLastFrame, target);
        }
    }
}
