/**
 *  @author Marian Ondrejicek (xondre02)
 *  @date   May 2012
 *
 *  This singleton holds some parameters, which may be useful in different
 *  AppStates. For example whether the game should run in splitscreen or if
 *  the effects are enabled. It also holds list of all players in the game.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "config.h"

#include "pugixml/pugixml.hpp"

// At the beginning, we have no instance
Config *Config::m_instance = NULL;

// Where the settings are stored
const std::string configFileName = "config.xml";

// Default nicknames if config file is not available
const std::string defaultNickSingle = "Player";
const std::string defaultNickSplit1 = "Player 1";
const std::string defaultNickSplit2 = "Player 2";

// Scoreboard displays only TOP 5 players
const unsigned int scoreboardSize = 5;

Config::Config()
{
    m_splitScreen = false;
    m_player = new Player;
    m_player1 = new Player;
    m_player2 = new Player;

    m_blur = true;
    m_glow = true;
    m_music = true;

    m_fs = new OgreBites::FileSystemLayerImpl("velocityoverflow");

    updatePlayers();
    load();
}

Config::~Config()
{
    delete m_player;
    delete m_player1;
    delete m_player2;
    delete m_fs;
}

void Config::load()
{
    pugi::xml_document doc;

    // If we couldn't load the file just use the defaults
    if (!doc.load_file(m_fs->getWritablePath(configFileName).c_str())) {
        m_player->setNick(defaultNickSingle);
        m_player1->setNick(defaultNickSplit1);
        m_player2->setNick(defaultNickSplit2);
        generateScoreboard();
        return;
    }

    pugi::xml_node xmlConfigNode = doc.child("config");

    pugi::xml_node xmlOptionsNode = xmlConfigNode.child("options");
    pugi::xml_node xmlBlurNode = xmlOptionsNode.child("blur");
    pugi::xml_node xmlGlowNode = xmlOptionsNode.child("glow");
    pugi::xml_node xmlMusicNode = xmlOptionsNode.child("music");

    pugi::xml_node xmlNicknamesNode = xmlConfigNode.child("nicknames");
    pugi::xml_node xmlPlayerNode = xmlNicknamesNode.child("player");
    pugi::xml_node xmlPlayer1Node = xmlNicknamesNode.child("player1");
    pugi::xml_node xmlPlayer2Node = xmlNicknamesNode.child("player2");

    m_blur = xmlBlurNode.attribute("enabled").as_bool();
    m_glow = xmlGlowNode.attribute("enabled").as_bool();
    m_music = xmlMusicNode.attribute("enabled").as_bool();

    m_player->setNick(xmlPlayerNode.attribute("name").value());
    m_player1->setNick(xmlPlayer1Node.attribute("name").value());
    m_player2->setNick(xmlPlayer2Node.attribute("name").value());

    // Load both scoreboards
    pugi::xml_node xmlScoreboardsNode = xmlConfigNode.child("scoreboards");
    pugi::xml_node sbNode = xmlScoreboardsNode.child("scoreboard");

    while (sbNode) {
        for (pugi::xml_node score = sbNode.child("score"); score; score = score.next_sibling()) {
            Score s;
            s.nick = score.attribute("nick").value();
            s.time = score.attribute("time").as_uint();

            if (std::string(sbNode.attribute("type").value()) == "single")
                m_scoreboardSingle.push_back(s);
            else
                m_scoreboardSplit.push_back(s);
        }

        sbNode = sbNode.next_sibling();
    }
}

void Config::save()
{
    pugi::xml_document doc;

    pugi::xml_node xmlConfigNode = doc.append_child("config");

    pugi::xml_node xmlOptionsNode = xmlConfigNode.append_child("options");
    pugi::xml_node xmlBlurNode = xmlOptionsNode.append_child("blur");
    pugi::xml_node xmlGlowNode = xmlOptionsNode.append_child("glow");
    pugi::xml_node xmlMusicNode = xmlOptionsNode.append_child("music");

    pugi::xml_node xmlNicknamesNode = xmlConfigNode.append_child("nicknames");
    pugi::xml_node xmlPlayerNode = xmlNicknamesNode.append_child("player");
    pugi::xml_node xmlPlayer1Node = xmlNicknamesNode.append_child("player1");
    pugi::xml_node xmlPlayer2Node = xmlNicknamesNode.append_child("player2");

    xmlBlurNode.append_attribute("enabled") = m_blur;
    xmlGlowNode.append_attribute("enabled") = m_glow;
    xmlMusicNode.append_attribute("enabled") = m_music;

    xmlPlayerNode.append_attribute("name") = m_player->getNick().c_str();
    xmlPlayer1Node.append_attribute("name") = m_player1->getNick().c_str();
    xmlPlayer2Node.append_attribute("name") = m_player2->getNick().c_str();

    // Save both scoreboards
    pugi::xml_node xmlScoreboardsNode = xmlConfigNode.append_child("scoreboards");

    pugi::xml_node sbSingleNode = xmlScoreboardsNode.append_child("scoreboard");
    sbSingleNode.append_attribute("type") = "single";

    pugi::xml_node sbSplitNode = xmlScoreboardsNode.append_child("scoreboard");
    sbSplitNode.append_attribute("type") = "split";

    for (int i = 0; i < m_scoreboardSingle.size(); i++) {
        pugi::xml_node score = sbSingleNode.append_child("score");
        score.append_attribute("nick") = m_scoreboardSingle[i].nick.c_str();
        score.append_attribute("time") = m_scoreboardSingle[i].time;
    }

    for (int i = 0; i < m_scoreboardSplit.size(); i++) {
        pugi::xml_node score = sbSplitNode.append_child("score");
        score.append_attribute("nick") = m_scoreboardSplit[i].nick.c_str();
        score.append_attribute("time") = m_scoreboardSplit[i].time;
    }

    doc.save_file(m_fs->getWritablePath(configFileName).c_str());
}

Config *Config::getSingleton()
{
    if (m_instance == NULL)
        m_instance = new Config;

    return m_instance;
}

void Config::updateScoreboard()
{
    if (m_splitScreen) {
        // Add latest scores only if players finished the race
        if (m_player1->getState() == PLAYER_STATE_FINISHED &&
            m_player2->getState() == PLAYER_STATE_FINISHED) {
            Score player1Score;
            player1Score.nick = m_player1->getNick();
            player1Score.time = m_player1->getTotalTime();

            Score player2Score;
            player2Score.nick = m_player2->getNick();
            player2Score.time = m_player2->getTotalTime();

            // Add latest scores to scoreboard
            m_scoreboardSplit.push_back(player1Score);
            m_scoreboardSplit.push_back(player2Score);
        }

        // Sort it by time
        std::sort(m_scoreboardSplit.begin(), m_scoreboardSplit.end());

        // Remove items at the bottom
        if (m_scoreboardSplit.size() > scoreboardSize)
            m_scoreboardSplit.resize(scoreboardSize);
    } else {
        // Add latest score only if player finished the race
        if (m_player->getState() == PLAYER_STATE_FINISHED) {
            Score playerScore;
            playerScore.nick = m_player->getNick();
            playerScore.time = m_player->getTotalTime();

            // Add latest score to scoreboard
            m_scoreboardSingle.push_back(playerScore);
        }

        // Sort it by time
        std::sort(m_scoreboardSingle.begin(), m_scoreboardSingle.end());

        // Remove items at the bottom
        if (m_scoreboardSingle.size() > scoreboardSize)
            m_scoreboardSingle.resize(scoreboardSize);
    }
}

Player *Config::getPlayer(int who)
{
    if (who == PLAYER_SPLIT1)
        return m_player1;
    else if (who == PLAYER_SPLIT2)
        return m_player2;
    else
        return m_player;
}

void Config::updatePlayers()
{
    m_players.clear();

    if (m_splitScreen) {
        m_players.push_back(m_player1);
        m_players.push_back(m_player2);
    } else {
        m_players.push_back(m_player);
    }
}

void Config::generateScoreboard()
{
    for (int i = 0; i < scoreboardSize; i++) {
        Score score;
        score.nick = "Bob";
        score.time = 100000 + i*5000;
        m_scoreboardSingle.push_back(score);
        m_scoreboardSplit.push_back(score);
    }
}
