/**
 *  @author Marian Ondrejicek (xondre02)
 *  @date   May 2012
 *
 *  This singleton holds some parameters, which may be useful in different
 *  AppStates. For example whether the game should run in splitscreen or if
 *  the effects are enabled. It also holds list of all players in the game.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef CONFIG_H
#define CONFIG_H

#include "player.h"

#include "OgreSDK/FileSystemLayerImpl.h"

#include <vector>

/** There are three types of players. */
enum PlayerType
{
    PLAYER_SINGLE,  // Player in single player mode
    PLAYER_SPLIT1,  // 1st player in split screen mode
    PLAYER_SPLIT2,  // 2nd player in split screen mode
};

/** One row of scoreboard. */
struct Score
{
    std::string nick;   // Player's nickname
    unsigned int time;  // Time in ms
    bool operator < (const Score &rhs) const { return time < rhs.time; }
};

/** Allows to share some useful data between different AppStates. */
class Config
{
protected:
    Config();
    ~Config();

    /** Load configuration from file when starting. */
    void load();

public:
    /** Save configuration to file. */
    void save();

    /** Get pointer to the actual object. */
    static Config *getSingleton();

    /** Enable/disable motion blur effect. */
    void setBlurEnabled(bool enabled) { m_blur = enabled; }

    /** Is motion blur effect enabled? */
    bool isBlurEnabled() const { return m_blur; }

    /** Enable/disable glow effect. */
    void setGlowEnabled(bool enabled) { m_glow = enabled; }

    /** Is glow effect enabled? */
    bool isGlowEnabled() const { return m_glow; }

    /** Enable/disable music. */
    void setMusicEnabled(bool enabled) { m_music = enabled; }

    /** Is music enabled? */
    bool isMusicEnabled() const { return m_music; }

    /** Recalculate scoreboard using current values.
     *  This should be called after every finished race. */
    void updateScoreboard();

    /** Return reference to the scoreboard for single player mode. */
    const std::vector<Score> &getScoreboardSingle() { return m_scoreboardSingle; }

    /** Return reference to the scoreboard for split screen mode. */
    const std::vector<Score> &getScoreboardSplit() { return m_scoreboardSplit; }

    /** Set to split screen mode or single player. */
    void setSplitScreen(bool enabled) { m_splitScreen = enabled; updatePlayers(); }

    /** Should we run in split screen mode? */
    bool isSplitScreen() const { return m_splitScreen; }

    /** Return number of screens (players) in the game. */
    int getScreensCount() const { return (m_splitScreen ? 2 : 1); }

    /** Return pointer to the Player object. */
    Player* getPlayer(int who=PLAYER_SINGLE);

    /** Return list of all players in the game. */
    std::vector<Player *> getPlayers() { return m_players; }

protected:
    /** (Re)create list of all players. */
    void updatePlayers();

    /** Generate default values for scoreboards if can't load file. */
    void generateScoreboard();

private:
    static Config *m_instance;  /**< The only instance of this class. */
    
    bool m_blur;                /**< Motion blur effect enabled? */
    bool m_glow;                /**< Glow effect enabled? */
    bool m_music;               /**< Music enabled? */

    bool m_splitScreen;         /**< Use split screen mode? */
    Player *m_player;           /**< Player in single player mode. */
    Player *m_player1;          /**< 1st player in split screen mode. */
    Player *m_player2;          /**< 2nd player in split screen mode. */

    /** This contains pointers to the Player objects above. This way
      * may be more practical, if we want to iterate over players.
      * Size of this list is number of players in the game. */
    std::vector<Player *> m_players;

    std::vector<Score> m_scoreboardSingle;
    std::vector<Score> m_scoreboardSplit;

    /** Platform independent path to user's folder. */
    OgreBites::FileSystemLayerImpl *m_fs;
};

#endif  // CONFIG_H
