/**
 *  @author Marian Ondrejicek (xondre02)
 *  @date   May 2012
 *
 *  Camera that dynamically follows the ship.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "dynamiccamera.h"

/** Maximal allowed value of time since last update() in miliseconds.
 *  This prevents camera from going crazy if the value is too high. */
const double lastFrameMax = 100;

template <class T>
T clamp(T value, T min, T max)
{
    if (value < min)
        return min;
    else if (value > max)
        return max;
    return value;
}

void DynamicCamera::setPosition(const Ogre::Vector3 &pos)
{
    m_desiredPos = pos;
    m_realPos = m_desiredPos;

    m_desiredLook = m_desiredPos;
    m_realLook = m_desiredLook;

    if (m_camera) {
        m_camera->setPosition(m_desiredPos);
        m_camera->lookAt(m_desiredLook);
    }
}

void DynamicCamera::update(double timeSinceLastFrame, const Ogre::SceneNode *target)
{
    // Just to make sure that camera doesn't go crazy
    if (timeSinceLastFrame > lastFrameMax)
        timeSinceLastFrame = lastFrameMax;

    updatePos(timeSinceLastFrame, target);
    updateLook(timeSinceLastFrame, target);
}

void DynamicCamera::updatePos(double timeSinceLastFrame, const Ogre::SceneNode *target)
{
    // Relative position of camera to the ship in idle state
    const Ogre::Vector3 desiredRelPos(0, 8, -15);

    // How much we can let the camera go away from desired position
    const Ogre::Real springLength = 2.0;

    // In case camera is too far, limit the maximal length to this value
    const Ogre::Real maxLength = 30.0;

    // Calculate current desired position for camera
    m_desiredPos = target->getPosition() + target->getOrientation() * desiredRelPos;

    Ogre::Real dist = m_desiredPos.distance(m_realPos);
    dist = clamp(dist, springLength, maxLength);

    Ogre::Real factor = dist / springLength;

    Ogre::Vector3 diff = m_desiredPos - m_realPos;
    m_realPos += (timeSinceLastFrame/1000.0) * factor * diff;

    m_camera->setPosition(m_realPos);
}

void DynamicCamera::updateLook(double timeSinceLastFrame, const Ogre::SceneNode *target)
{
    // Offset from center of the ship where the camera will be looking at
    const Ogre::Vector3 desiredRelLook(0, 0, 20);

    // How much we can let the look at position go away from desired
    const Ogre::Real springLength = 2.0;

    // In case camera is too far, limit the maximal length to this value
    const Ogre::Real maxLength = 30.0;

    // Calculate current desired look at position for camera
    m_desiredLook = target->getPosition() + target->getOrientation() * desiredRelLook;

    Ogre::Real dist = m_desiredLook.distance(m_realLook);
    dist = clamp(dist, springLength, maxLength);

    Ogre::Real factor = dist / springLength;

    Ogre::Vector3 diff = m_desiredLook - m_realLook;
    m_realLook += (timeSinceLastFrame/1000.0) * factor * diff;

    m_camera->lookAt(m_realLook);
}
