/**
 *  @author Marian Ondrejicek (xondre02)
 *  @date   May 2012
 *
 *  Manages and connects essential parts of the game.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "gamestate.h"
#include "physics.h"
#include "levelobjects.h"
#include "soundmanager.h"
#include "config.h"

// Z-order of the master viewport must be higher than all others
const int masterViewportZorder = 100;

// Offset (in pixels) of HUD elements from the screen edges
const int hudOffset = 10;

// Higher value means less motion blur at certain speed
const float blurFactor = 400;

// How many laps player has to go to finish the race
const int raceLaps = 2;

// How many ms after entering the game will the race start
const int startCountdown = 4500;

// How many ms after finished race the results will be shown
const int resultsCountdown = 1500;

GameState::GameState()
{
    // Only glow effect listener is created just once.
    // Other stuff is recreated with every gameplay.
    m_glowListener = new GlowListener;
    Ogre::MaterialManager::getSingleton().addListener(m_glowListener);

    m_timer = new RaceTimer;
}

GameState::~GameState()
{
    delete m_glowListener;
    delete m_timer;
}

void GameState::enter()
{
    // Create scene manager for this state
    m_sceneManager = OgreFramework::getSingletonPtr()->m_root->
        createSceneManager(Ogre::ST_GENERIC, "GameSceneManager");

    // Create manager for all in-game cameras
    m_camManager = new CameraManager(m_sceneManager);

    // Connect the scene manager with sound manager
    SoundManager::getSingleton()->attach(m_sceneManager);

    for (int i = 0; i < Config::getSingleton()->getScreensCount(); i++) {
        Player *player = Config::getSingleton()->getPlayers()[i];
        player->init();
        player->setState(PLAYER_STATE_WAITING);

        m_ships.push_back(new Ship);
        m_blurListeners.push_back(new BlurListener);
        
        // Create new viewport. Note: This code works only with one or two viewports!
        Ogre::Viewport *viewport;
        if (Config::getSingleton()->isSplitScreen()) {
            // Create 1st viewport on the right and 2nd on the left
            viewport = OgreFramework::getSingletonPtr()->m_renderWindow->
                addViewport(NULL, i, (i ? 0.0 : 0.5), 0.0, 0.5, 1.0);
        } else {
            // Create one big viewport over the whole screen
            viewport = OgreFramework::getSingletonPtr()->m_renderWindow->
                addViewport(NULL, i);
        }
        viewport->setBackgroundColour(Ogre::ColourValue(0.0, 0.0, 0.0));

        // Create camera behind the ship and add it to the viewport
        Ogre::Real ratio = (Ogre::Real)viewport->getActualWidth() / viewport->getActualHeight();
        GameCamera *gameCam = m_camManager->createCamera(CAMERA_DYNAMIC, ratio, i);
        viewport->setCamera(gameCam->getCamera());

        // Create rear camera looking at the ship
        m_camManager->createCamera(CAMERA_SIDE, ratio, i);

        // Save viewport
        m_viewports.push_back(viewport);
    }

    // Create camera and transparent viewport, which covers entire screen (used for GUI)
    m_masterViewport = OgreFramework::getSingletonPtr()->m_renderWindow->
        addViewport(NULL, masterViewportZorder);
    m_masterViewport->setClearEveryFrame(true, Ogre::FBT_DEPTH);
    m_masterViewport->setVisibilityMask(0);
    m_masterCamera = m_sceneManager->createCamera("MasterViewportCamera");
    m_masterCamera->setAspectRatio(Ogre::Real(m_masterViewport->getActualWidth())/
        Ogre::Real(m_masterViewport->getActualHeight()));
    m_masterViewport->setCamera(m_masterCamera);

    // Get index of last added viewport, the master viewport
    int viewportIndex = OgreFramework::getSingletonPtr()->m_renderWindow->getNumViewports() - 1;

    // Associate GUI with the scene
    Gui::getSingleton()->attach(m_sceneManager);
    Gui::getSingleton()->getPlatform()->getRenderManagerPtr()->setActiveViewport(viewportIndex);

    createScene();
    createGui();
    resetHud();

    m_raceStarted = false;
    m_raceFinished = false;
    m_resultsShown = false;
    m_startCountdown = startCountdown;
    m_resultsCountdown = 0;

    m_timer->start();

    m_paused = false;
}

void GameState::exit()
{
    // Disconnect the GUI
    Gui::getSingleton()->detach();

    // Stop the physical simulation
    Physics::getSingleton()->destroy();

    // Destroy level
    Level::getSingleton()->destroy();

    for (int i = 0; i < Config::getSingleton()->getScreensCount(); i++) {
        // Disable compositor effects
        Ogre::CompositorManager::getSingleton().setCompositorEnabled(m_viewports[i], "Glow", false);
        Ogre::CompositorManager::getSingleton().removeCompositor(m_viewports[i], "Glow");
        Ogre::CompositorManager::getSingleton().setCompositorEnabled(m_viewports[i], "RadialBlur", false);
        Ogre::CompositorManager::getSingleton().removeCompositor(m_viewports[i], "RadialBlur");
        delete m_blurListeners[i];

        // Destroy viewport
        OgreFramework::getSingleton().m_renderWindow->removeViewport(i);

        delete m_ships[i];
    }

    // Destroy all in-game cameras
    delete m_camManager;

    // Destroy master viewport and associated camera
    OgreFramework::getSingleton().m_renderWindow->removeViewport(masterViewportZorder);
    m_sceneManager->destroyCamera(m_masterCamera);

    // The lists need to be reseted too, so next game starts clean
    m_blurListeners.clear();
    m_viewports.clear();
    m_ships.clear();

    // Disconnect the sound manager
    SoundManager::getSingleton()->detach();

    // Destroy the scene manager
    if (m_sceneManager)
        OgreFramework::getSingletonPtr()->m_root->destroySceneManager(m_sceneManager);
}

bool GameState::keyPressed(const OIS::KeyEvent &arg)
{
    OIS::Keyboard *keyboard = OgreFramework::getSingletonPtr()->m_keyboard;

    // Pause the game?
    if (keyboard->isKeyDown(OIS::KC_ESCAPE) && !m_raceFinished) {
        if (m_paused) {
            resumeGame();
        } else {
            pauseGame();
        }
        return true;
    }

    // First ship
    if (Config::getSingleton()->getPlayers()[0]->getState() == PLAYER_STATE_RACING) {
        if (keyboard->isKeyDown(OIS::KC_UP)) {
            m_ships[0]->accelerate(true);
        }
        if (keyboard->isKeyDown(OIS::KC_DOWN)) {
            m_ships[0]->slowDown(true);
        }
        if (keyboard->isKeyDown(OIS::KC_LEFT)) {
            m_ships[0]->steerLeft(true);
        }
        if (keyboard->isKeyDown(OIS::KC_RIGHT)) {
            m_ships[0]->steerRight(true);
        }

        if (keyboard->isKeyDown(OIS::KC_HOME)) {
            resetShip(0);
        }
    }

    // Second ship
    if (Config::getSingleton()->isSplitScreen()) {
        if (Config::getSingleton()->getPlayers()[1]->getState() == PLAYER_STATE_RACING) {
            if (keyboard->isKeyDown(OIS::KC_W)) {
                m_ships[1]->accelerate(true);
            }
            if (keyboard->isKeyDown(OIS::KC_S)) {
                m_ships[1]->slowDown(true);
            }
            if (keyboard->isKeyDown(OIS::KC_A)) {
                m_ships[1]->steerLeft(true);
            }
            if (keyboard->isKeyDown(OIS::KC_D)) {
                m_ships[1]->steerRight(true);
            }

            if (keyboard->isKeyDown(OIS::KC_R)) {
                resetShip(1);
            }
        }
    }

#ifdef GAME_DEBUG

    // Use dynamic camera behind the ship
    if (keyboard->isKeyDown(OIS::KC_PGUP)) {
        Ogre::Camera *cam = m_camManager->getCamera(CAMERA_DYNAMIC, 0)->getCamera();
        changeCamera(m_viewports[0], cam, CAMERA_DYNAMIC);
        setHudVisible(0, true);
    }

    // User camera looking at the side of the ship
    if (keyboard->isKeyDown(OIS::KC_PGDOWN)) {
        Ogre::Camera *cam = m_camManager->getCamera(CAMERA_SIDE, 0)->getCamera();
        changeCamera(m_viewports[0], cam, CAMERA_SIDE);
        setHudVisible(0, false);
    }

    int camKey = -1;
    if (keyboard->isKeyDown(OIS::KC_0)) camKey = 0;
    if (keyboard->isKeyDown(OIS::KC_1)) camKey = 1;
    if (keyboard->isKeyDown(OIS::KC_2)) camKey = 2;
    if (keyboard->isKeyDown(OIS::KC_3)) camKey = 3;
    if (keyboard->isKeyDown(OIS::KC_4)) camKey = 4;
    if (keyboard->isKeyDown(OIS::KC_5)) camKey = 5;
    if (keyboard->isKeyDown(OIS::KC_6)) camKey = 6;
    if (keyboard->isKeyDown(OIS::KC_7)) camKey = 7;
    if (keyboard->isKeyDown(OIS::KC_8)) camKey = 8;
    if (keyboard->isKeyDown(OIS::KC_9)) camKey = 9;

    // Switch to one of the static cameras?
    if (camKey >= 0) {
        int camerasCount = Level::getSingleton()->getCameras()->size();
        if (camerasCount > camKey) {
            Ogre::Camera *cam = m_camManager->getCamera(CAMERA_STATIC, 0, camKey)->getCamera();
            changeCamera(m_viewports[0], cam, CAMERA_STATIC);
            setHudVisible(0, false);
        }
    }

#endif

    // Toggle physics debug mode
    if (keyboard->isKeyDown(OIS::KC_F1)) {
        Physics::getSingleton()->toggleDebug();
    }

    // Physics showcase
    if (keyboard->isKeyDown(OIS::KC_G)) {
        physicsDemo();
    }

    // Forward events to the framework
    OgreFramework::getSingletonPtr()->keyPressed(arg);
    return true;
}

bool GameState::keyReleased(const OIS::KeyEvent &arg)
{
    // First ship
    if (arg.key == OIS::KC_UP) {
        m_ships[0]->accelerate(false);
    }
    if (arg.key == OIS::KC_DOWN) {
        m_ships[0]->slowDown(false);
    }
    if (arg.key == OIS::KC_LEFT) {
        m_ships[0]->steerLeft(false);
    }
    if (arg.key == OIS::KC_RIGHT) {
        m_ships[0]->steerRight(false);
    }

    // Second ship
    if (Config::getSingleton()->isSplitScreen()) {
        if (arg.key == OIS::KC_W) {
            m_ships[1]->accelerate(false);
        }
        if (arg.key == OIS::KC_S) {
            m_ships[1]->slowDown(false);
        }
        if (arg.key == OIS::KC_A) {
            m_ships[1]->steerLeft(false);
        }
        if (arg.key == OIS::KC_D) {
            m_ships[1]->steerRight(false);
        }
    }

    // Forward events to the framework
    OgreFramework::getSingletonPtr()->keyReleased(arg);
    return true;
}

bool GameState::mouseMoved(const OIS::MouseEvent &arg)
{
    MyGUI::InputManager::getInstance().injectMouseMove(arg.state.X.abs, arg.state.Y.abs, arg.state.Z.abs);
    return true;
}

bool GameState::mousePressed(const OIS::MouseEvent &arg, OIS::MouseButtonID id)
{
    MyGUI::InputManager::getInstance().injectMousePress(arg.state.X.abs, arg.state.Y.abs, MyGUI::MouseButton::Enum(id));
    return true;
}

bool GameState::mouseReleased(const OIS::MouseEvent &arg, OIS::MouseButtonID id)
{
    MyGUI::InputManager::getInstance().injectMouseRelease(arg.state.X.abs, arg.state.Y.abs, MyGUI::MouseButton::Enum(id));
    return true;
}

void GameState::guiRestartPressed(MyGUI::Widget *)
{
    Config::getSingleton()->save();
    exit();
    enter();
}

void GameState::guiResumePressed(MyGUI::Widget *)
{
    resumeGame();
}

void GameState::guiExitPressed(MyGUI::Widget *)
{
    Config::getSingleton()->save();
    changeAppState(findByName("MenuState"));
}

void GameState::pauseGame()
{
    MyGUI::PointerManager::getInstance().setVisible(true);
    Gui::getSingleton()->getGui()->findWidget<MyGUI::Widget>("PauseBackPanel")->setVisible(true);
    m_timer->pause();
    m_paused = true;
}

void GameState::resumeGame()
{
    MyGUI::PointerManager::getInstance().setVisible(false);
    Gui::getSingleton()->getGui()->findWidget<MyGUI::Widget>("PauseBackPanel")->setVisible(false);
    m_timer->resume();
    m_paused = false;
}

void GameState::update(double timeSinceLastFrame)
{
    // Show FPS
    //OutputDebugString(Ogre::StringConverter::toString(Ogre::Real(1000.0/timeSinceLastFrame)).c_str());
    //OutputDebugString("\n");
    
    // If the song stopped playing, play another one
    if (SoundManager::getSingleton()->isSongStopped())
        SoundManager::getSingleton()->playRandom();
    
    for (int i = 0; i < Config::getSingleton()->getScreensCount(); i++) {
        m_ships[i]->update(timeSinceLastFrame, m_paused);
        
        if (m_paused)
            continue;

        // Update position and orientation of all cameras
        m_camManager->update(timeSinceLastFrame, m_ships[i]->getNode(), i);

        // Update motion blur according to current speed
        Ogre::Real strength = m_ships[i]->getSpeed() / blurFactor;
        m_blurListeners[i]->setSampleStrength(strength);
    }

    if (m_paused)
        return;

    // Start the race after initial countdown finishes
    m_startCountdown -= timeSinceLastFrame;
    if (!m_raceStarted &&  m_startCountdown <= 0) {
        // Activate each player
        for (int i = 0; i < Config::getSingleton()->getScreensCount(); i++) {
            Player *player = Config::getSingleton()->getPlayers()[i];
            player->setState(PLAYER_STATE_RACING);
        }

        // Simulate keyboard press (ugly way to make sure that player will accelerate
        // if he has pressed the accelerate key before race started).
        keyPressed(OIS::KeyEvent(NULL, OIS::KC_EQUALS, 0));

        m_timer->start();
        m_raceStarted = true;
    }

    // Update physics (stepSimulation takes time in seconds, not ms)
    const int maxSteps = 10;	// Maximum number of steps for one call
    Physics::getSingleton()->getWorld()->stepSimulation(timeSinceLastFrame/1000.0, maxSteps);
    Physics::getSingleton()->getDbgdraw()->step();

    // How many players has finished the game
    int playersFinished = 0;

    // Check each player
    for (int i = 0; i < Config::getSingleton()->getScreensCount(); i++) {
        // Pointer to current player
        Player *player = Config::getSingleton()->getPlayers()[i];

        // If the player is in space place him back on track
        if (m_ships[i]->isInSpace()) {
            resetShip(i);
        }

        // If the player goes wrong way reset him to last checkpoint
        if (checkWrongWay(player)) {
            player->setWrongWay(false);
            resetShip(i);
        }

        // Has the player finished the race?
        if (player->getState() == PLAYER_STATE_RACING && player->getLaps() > raceLaps) {
            // Stop the ship
            m_ships[i]->accelerate(false);
            m_ships[i]->slowDown(false);
            m_ships[i]->steerLeft(false);
            m_ships[i]->steerRight(false);

            // Change camera
            Ogre::Camera *staticCamera = m_camManager->getCamera(CAMERA_STATIC, i, 0)->getCamera();
            changeCamera(m_viewports[i], staticCamera, CAMERA_STATIC);
            setHudVisible(i, false);

            // Remember total race time
            player->setTotalTime(m_timer->getMilliseconds());

            // Set new state
            player->setState(PLAYER_STATE_FINISHED);
        }

        if (player->getState() == PLAYER_STATE_FINISHED) {
            playersFinished++;
        }
    }

    // If all players finished the race end the game
    if (playersFinished == Config::getSingleton()->getScreensCount()) {
        // Do this just once
        if (!m_raceFinished) {
            m_resultsCountdown = resultsCountdown;
            m_raceFinished = true;
        }
    }

    if (m_raceFinished && !m_resultsShown) {
        if (m_resultsCountdown <= 0) {
            // Show the results screen
            m_resultsShown = true;
            showResults();

            // Recalculate total scoreboard (shown in menu)
            Config::getSingleton()->updateScoreboard();
        } else {
            // Decrease the countdown to show the results later
            m_resultsCountdown -= timeSinceLastFrame;
        }
    }

    // Update HUD values
    updateHud();
}

void GameState::createScene()
{
    // Setup shadows
    m_sceneManager->setShadowTechnique(Ogre::SHADOWTYPE_STENCIL_MODULATIVE);
    m_sceneManager->setShadowColour(Ogre::ColourValue(0.5, 0.5, 0.5));

    // Create level
    Level::getSingleton()->setSceneManager(m_sceneManager);
    Level::getSingleton()->create("city");

    // Aspect ratio of level cameras
    Ogre::Real ratio = (Ogre::Real)m_viewports[0]->getActualWidth() / m_viewports[0]->getActualHeight();

    // Create cameras for this level for each player
    const std::vector<Ogre::Vector3> *levelCameras = Level::getSingleton()->getCameras();
    for (unsigned int i = 0; i < levelCameras->size(); i++) {
        for (unsigned int j = 0; j < m_ships.size(); j++) {
            const Ogre::Vector3 &cameraPos = (*levelCameras)[i];
            m_camManager->createCamera(CAMERA_STATIC, ratio, j)->setPosition(cameraPos);
        }
    }

    for (int i = 0; i < Config::getSingleton()->getScreensCount(); i++) {
        // Get starting position and rotation of current ship
        Location dock = Level::getSingleton()->getDock(i);

        // Pointer to the player
        Player *player = Config::getSingleton()->getPlayers()[i];

        // Find out what ship was selected in menu for current player
        std::string shipName = player->getShipName();

        // Let the player know where the track has start and half checkpoints,
        // so he can make correct computations about traveled laps.
        player->setFirstCheckpoint(Level::getSingleton()->getFirstCheckpoint());
        player->setHalfCheckpoint(Level::getSingleton()->getHalfCheckpoint());

        // Create ship
        m_ships[i]->setSceneManager(m_sceneManager);
        m_ships[i]->create(shipName, dock);
        m_ships[i]->initPhysics();
        player->setShip(m_ships[i]);

        // Set camera's position
        m_camManager->getCamera(CAMERA_DYNAMIC, i)->setPosition(dock.position);
    
        // Add ship to the physics world
        Physics::getSingleton()->addShip(m_ships[i], player);

        // Pointer to the viewport of current player
        Ogre::Viewport *viewport = m_viewports[i];

        // Enable glow effect?
        if (Config::getSingleton()->isGlowEnabled()) {
            Ogre::CompositorManager::getSingleton().addCompositor(viewport, "Glow");
            Ogre::CompositorManager::getSingleton().setCompositorEnabled(viewport, "Glow", true);
        }

        // Enable motion blur effect?
        if (Config::getSingleton()->isBlurEnabled()) {
            Ogre::CompositorManager::getSingleton().addCompositor(viewport, "RadialBlur");
            Ogre::CompositorManager::getSingleton().setCompositorEnabled(viewport, "RadialBlur", true);
            Ogre::CompositorInstance *compositorInstance = Ogre::CompositorManager::getSingleton().
                getCompositorChain(viewport)->getCompositor("RadialBlur");
            compositorInstance->addListener(m_blurListeners[i]);
        }
    }

    // Setup sounds
    SoundManager::getSingleton()->createGameSounds();
    SoundManager::getSingleton()->playRandom();
}

void GameState::createGui()
{
    // Load layout from file
    Gui::getSingleton()->load("GameState.layout");

    MyGUI::Gui *myGui = Gui::getSingleton()->getGui();

    // Register delegates for buttons in pause menu
    MyGUI::ButtonPtr resumeButton = myGui->findWidget<MyGUI::Button>("ResumeButton");
    resumeButton->eventMouseButtonClick += MyGUI::newDelegate(this, &GameState::guiResumePressed);
    MyGUI::ButtonPtr exitButton = myGui->findWidget<MyGUI::Button>("ExitButton");
    exitButton->eventMouseButtonClick += MyGUI::newDelegate(this, &GameState::guiExitPressed);

    // Size of the screen in pixels
    int width = m_masterViewport->getActualWidth() / Config::getSingleton()->getScreensCount();
    int height = m_masterViewport->getActualHeight();
    int hudWidth = width - 2*hudOffset;

    // Adjust sizes and positions of HUD elements to fit the screen of each player
    for (int i = 0; i < Config::getSingleton()->getScreensCount(); i++) {
        // HUD elements of each player have different names (ending with 0 or 1)
        std::string id = Ogre::StringConverter::toString(i);

        // X position of HUD panels of current player (count-i-1 because 1st player is on the right)
        int hudX = (Config::getSingleton()->getScreensCount()-i-1)*width + hudOffset;

        // Adjust top panel
        MyGUI::WidgetPtr hudPanelTop = myGui->findWidget<MyGUI::Widget>("HudPanelTop" + id);
        hudPanelTop->setPosition(hudX, 0);
        hudPanelTop->setSize(hudWidth, hudPanelTop->getHeight());

        // Adjust bottom panel
        MyGUI::WidgetPtr hudPanelBottom = myGui->findWidget<MyGUI::Widget>("HudPanelBottom" + id);
        hudPanelBottom->setPosition(hudX, height - hudPanelBottom->getHeight());
        hudPanelBottom->setSize(hudWidth, hudPanelBottom->getHeight());
    }

    // Disable 2nd player's HUD if not in split screen mode
    if (!Config::getSingleton()->isSplitScreen()) {
        myGui->findWidget<MyGUI::Widget>("HudPanelTop1")->setVisible(false);
        myGui->findWidget<MyGUI::Widget>("HudPanelBottom1")->setVisible(false);
    }

    // Hide the GUI and also the mouse cursor
    myGui->findWidget<MyGUI::Widget>("PauseBackPanel")->setVisible(false);
    MyGUI::PointerManager::getInstance().setVisible(false);
}

void GameState::showResults()
{
    // Load layout from file
    Gui::getSingleton()->load("GameStateResults.layout");

    MyGUI::Gui *myGui = Gui::getSingleton()->getGui();

    // Register delegates for buttons in results screen
    MyGUI::ButtonPtr restartButton = myGui->findWidget<MyGUI::Button>("RestartButton");
    restartButton->eventMouseButtonClick += MyGUI::newDelegate(this, &GameState::guiRestartPressed);
    MyGUI::ButtonPtr exitButton = myGui->findWidget<MyGUI::Button>("ExitButton");
    exitButton->eventMouseButtonClick += MyGUI::newDelegate(this, &GameState::guiExitPressed);

    // Fill the result values
    if (Config::getSingleton()->isSplitScreen()) {
        const Player *player1 = Config::getSingleton()->getPlayer(PLAYER_SPLIT1);
        const Player *player2 = Config::getSingleton()->getPlayer(PLAYER_SPLIT2);
        const std::string &player1time = RaceTimer::timeToString(player1->getTotalTime(), true);
        const std::string &player2time = RaceTimer::timeToString(player2->getTotalTime(), true);

        if (player1->getTotalTime() < player2->getTotalTime()) {
            // Player1 is the winner
            myGui->findWidget<MyGUI::TextBox>("FirstPlayer")->setCaption("1. " + player1->getNick());
            myGui->findWidget<MyGUI::TextBox>("FirstTime")->setCaption(player1time);
            myGui->findWidget<MyGUI::TextBox>("SecondPlayer")->setCaption("2. " + player2->getNick());
            myGui->findWidget<MyGUI::TextBox>("SecondTime")->setCaption(player2time);
        } else {
            // Player2 is the winner
            myGui->findWidget<MyGUI::TextBox>("FirstPlayer")->setCaption("1. " + player2->getNick());
            myGui->findWidget<MyGUI::TextBox>("FirstTime")->setCaption(player2time);
            myGui->findWidget<MyGUI::TextBox>("SecondPlayer")->setCaption("2. " + player1->getNick());
            myGui->findWidget<MyGUI::TextBox>("SecondTime")->setCaption(player1time);
        }
    } else {
        const Player *player = Config::getSingleton()->getPlayer();
        const std::string &time = RaceTimer::timeToString(player->getTotalTime(), true);

        myGui->findWidget<MyGUI::TextBox>("FirstPlayer")->setCaption("1. " + player->getNick());
        myGui->findWidget<MyGUI::TextBox>("FirstTime")->setCaption(time);
        myGui->findWidget<MyGUI::TextBox>("SecondPlayer")->setCaption("");
        myGui->findWidget<MyGUI::TextBox>("SecondTime")->setCaption("");
    }

    // Show the mouse cursor
    MyGUI::PointerManager::getInstance().setVisible(true);
}

void GameState::resetHud()
{
    MyGUI::Gui *myGui = Gui::getSingleton()->getGui();

    // Reset HUD values of each player to defaults
    for (int i = 0; i < Config::getSingleton()->getScreensCount(); i++) {
        // HUD elements of each player have different names (ending with 0 or 1)
        std::string id = Ogre::StringConverter::toString(i);
        
        // Find out current player's name
        std::string playerName = Config::getSingleton()->getPlayers()[i]->getNick();

        myGui->findWidget<MyGUI::TextBox>("PlayerNameBox" + id)->setCaption(playerName);
        myGui->findWidget<MyGUI::TextBox>("LapBox" + id)->setCaption("1/2");
        myGui->findWidget<MyGUI::TextBox>("TimeBox" + id)->setCaption("0:00:0");
        myGui->findWidget<MyGUI::TextBox>("SpeedBox" + id)->setCaption("0");
    }
}

void GameState::updateHud()
{
    MyGUI::Gui *myGui = Gui::getSingleton()->getGui();

    // We can't update HUD after finished race (it will be replaced with results screen)
    if (m_raceFinished) {
        return;
    }

    // If the race has not started update only the initial countdown
    if (!m_raceStarted) {
        int countValue = ceil(m_startCountdown / 1000.0);
        std::string countText = Ogre::StringConverter::toString(countValue);
        myGui->findWidget<MyGUI::TextBox>("StartBox")->setCaption(countText);
        return;
    }

    // Display or hide the GO! message
    if (m_startCountdown > -1000) {
        myGui->findWidget<MyGUI::TextBox>("StartBox")->setCaption("GO!");
    } else {
        myGui->findWidget<MyGUI::Widget>("CountdownPanel")->setVisible(false);
    }

    // Update HUD values of each player
    for (int i = 0; i < Config::getSingleton()->getScreensCount(); i++) {
        // Pointer to the player
        Player *player = Config::getSingleton()->getPlayers()[i];

        // No need to update if player has finished the race
        if (player->getState() == PLAYER_STATE_FINISHED) {
            continue;
        }

        // HUD elements of each player have different names (ending with 0 or 1)
        std::string id = Ogre::StringConverter::toString(i);
        
        // Update displayed laps
        std::string laps = Ogre::StringConverter::toString(player->getLaps());
        std::string totalLaps = Ogre::StringConverter::toString(raceLaps);
        myGui->findWidget<MyGUI::TextBox>("LapBox" + id)->setCaption(laps + "/" + totalLaps);

        // Update displayed speed
        std::string speed = Ogre::StringConverter::toString((int)m_ships[i]->getSpeed());
        myGui->findWidget<MyGUI::TextBox>("SpeedBox" + id)->setCaption(speed);

        // Update displayed time
        const std::string &time = RaceTimer::timeToString(m_timer->getMilliseconds());
        myGui->findWidget<MyGUI::TextBox>("TimeBox" + id)->setCaption(time);
    }
}

void GameState::setHudVisible(int player, bool visible)
{
    MyGUI::Gui *myGui = Gui::getSingleton()->getGui();

    if (player == 0) {
        myGui->findWidget<MyGUI::Widget>("HudPanelTop0")->setVisible(visible);
        myGui->findWidget<MyGUI::Widget>("HudPanelBottom0")->setVisible(visible);
    }
    else if (player == 1 && Config::getSingleton()->isSplitScreen()){
        myGui->findWidget<MyGUI::Widget>("HudPanelTop1")->setVisible(visible);
        myGui->findWidget<MyGUI::Widget>("HudPanelBottom1")->setVisible(visible);
    }
}

void GameState::resetShip(int index)
{
    // Get player associated with this ship
    Player *player = Config::getSingleton()->getPlayers()[index];

    // Get ID of current checkpoint
    unsigned int cp = player->getCurrentCheckpoint();

    // Find location of that checkpoint
    Location loc = Level::getSingleton()->getCheckpoint(cp);

    // Reset the ship and the camera to that location
    m_ships[index]->setLocation(loc);
    m_camManager->getCamera(CAMERA_DYNAMIC, index)->setPosition(loc.position);
}

void GameState::changeCamera(Ogre::Viewport *viewport, Ogre::Camera *cam, int cameraType)
{
    // This method assumes that compositors are already initialized

    // At first set the camera in the viewport
    viewport->setCamera(cam);

    // Disable compositor effects
    Ogre::CompositorManager::getSingleton().setCompositorEnabled(viewport, "Glow", false);
    Ogre::CompositorManager::getSingleton().setCompositorEnabled(viewport, "RadialBlur", false);

    // Re-enable glow effect?
    if (Config::getSingleton()->isGlowEnabled()) {
        Ogre::CompositorManager::getSingleton().setCompositorEnabled(viewport, "Glow", true);
    }

    // Re-enable blur effect?
    if (Config::getSingleton()->isBlurEnabled() && cameraType == CAMERA_DYNAMIC) {
        Ogre::CompositorManager::getSingleton().setCompositorEnabled(viewport, "RadialBlur", true);
    }
}

void GameState::physicsDemo()
{
    for (int i = 0; i < 1; ++i) {
        Ogre::SceneNode *boxNode = m_sceneManager->createSceneNode();
        Ogre::Vector3 pos = m_ships[0]->getNode()->getPosition();
        pos.y += 50*(i+1);
        boxNode->setPosition(pos);

        Ogre::Entity *boxEntity = m_sceneManager->createEntity("box.mesh");
        m_sceneManager->getRootSceneNode()->addChild(boxNode);
        boxNode->scale(2, 2, 2);
        boxNode->attachObject(boxEntity);

        // Create shape
        BtOgre::StaticMeshToShapeConverter converter(boxEntity);
        btCollisionShape *shape = converter.createBox();

        // Calculate inertia
        btScalar mass = 1000;
        btVector3 inertia;
        shape->calculateLocalInertia(mass, inertia);

        // Create MotionState
        BtOgre::RigidBodyState *state = new BtOgre::RigidBodyState(boxNode);

        // Create body
        btRigidBody *body = new btRigidBody(mass, state, shape, inertia);
        //body->setFriction(5.0);

        // Add it to physics world
        Physics::getSingleton()->getWorld()->addRigidBody(body);
    }
}

bool GameState::checkWrongWay(Player *player)
{
    unsigned int currentCp = player->getCurrentCheckpoint();

    // If we reach the next checkpoint, calculate new one
    if (currentCp == player->getNextCheckpoint()) {
        player->setNextCheckpoint(Level::getSingleton()->getNextCheckpoint(currentCp));
        player->setWrongWay(false);
    }

    // Previous of next should be current... otherwise we are going wrong way
    if (currentCp != Level::getSingleton()->getPrevCheckpoint(player->getNextCheckpoint())) {
        player->setNextCheckpoint(Level::getSingleton()->getNextCheckpoint(currentCp));
        player->setWrongWay(true);
    }

    return player->isWrongWay();
}
