/**
 *  @author Marian Ondrejicek (xondre02)
 *  @date   May 2012
 *
 *  Manages and connects essential parts of the game.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef GAMESTATE_H
#define GAMESTATE_H

#include "OgreFramework/appstate.h"
#include "cameramanager.h"
#include "level.h"
#include "ship.h"
#include "racetimer.h"
#include "glowlistener.h"
#include "blurlistener.h"
#include "gui.h"

/** This state is active during gameplay. */
class GameState : public AppState
{
public:
    GameState();
    ~GameState();

    DECLARE_APPSTATE_CLASS(GameState)

    void enter();
    void exit();

    bool keyPressed(const OIS::KeyEvent &arg);
    bool keyReleased(const OIS::KeyEvent &arg);

    bool mouseMoved(const OIS::MouseEvent &arg);
    bool mousePressed(const OIS::MouseEvent &arg, OIS::MouseButtonID id);
    bool mouseReleased(const OIS::MouseEvent &arg, OIS::MouseButtonID id);

    void guiRestartPressed(MyGUI::Widget *);
    void guiResumePressed(MyGUI::Widget *);
    void guiExitPressed(MyGUI::Widget *);

    /** Pause gameplay and display menu. */
    void pauseGame();

    /** Resume gameplay and hide menu. */
    void resumeGame();

    /** Update camera, ship, physics, etc. */
    void update(double timeSinceLastFrame);

    /** Place objects into the scene. */
    void createScene();

    /** Load GUI elements. */
    void createGui();

    /** Create and show screen with race results. */
    void showResults();

    /** Reset HUD (Head Up Display) of each player (name, speed, etc). */
    void resetHud();

    /** Update HUD values (speed, time, etc). */
    void updateHud();

    /** Show/hide HUD of specific player. */
    void setHudVisible(int player, bool visible);

    /** Reset ship to last checkpoint. */
    void resetShip(int index);

    /** Change camera of given viewport and update compositors. */
    void changeCamera(Ogre::Viewport *viewport, Ogre::Camera *cam, int cameraType);

    void physicsDemo();

protected:
    /** Check if player goes wrong way. Modifies the Player object
      * and also returns true if he does go wrong way. */
    bool checkWrongWay(Player *player);

private:
    Ogre::SceneManager *m_sceneManager;

    Ogre::Viewport *m_masterViewport;               /**< Viewport used for GUI rendering. */
    Ogre::Camera *m_masterCamera;                   /**< Camera used with master viewport. */
    
    CameraManager *m_camManager;                    /**< Manages all in-game cameras. */

    std::vector<Ogre::Viewport *> m_viewports;      /**< List of all viewports. */
    std::vector<Ship *> m_ships;                    /**< List of all ships in the race. */

    std::vector<BlurListener *> m_blurListeners;    /**< Blur effect listeners for all viewports. */
    GlowListener *m_glowListener;                   /**< Glow effect listener. */

    bool m_raceStarted;     /**< True if race is on (after initial countdown). */
    bool m_raceFinished;    /**< True if all players finished the race. */
    bool m_resultsShown;    /**< True if results screen was shown. */

    /** How much time left (in ms) to start the race. */
    int m_startCountdown;

    /** How much time left (in ms) to show the results after finished race. */
    int m_resultsCountdown;

    RaceTimer *m_timer;
    bool m_paused;
};

#endif  // GAMESTATE_H
