/**
 *  @author Marian Ondrejicek (xondre02)
 *  @date   May 2012
 *
 *  Class that represents level of the game.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "level.h"

#include "pugixml/pugixml.hpp"

// At the beginning, we have no instance
Level *Level::m_instance = NULL;

// Path to directory with level files
#ifdef GAME_LEVELS_DIR
const std::string levelDir = GAME_LEVELS_DIR;
#else
const std::string levelDir = "../media/levels/";
#endif

// Extension of level file
const std::string levelExt = ".level";

// Track needs to be tagged so it can be distinguished from other objects during ray tests
static int trackTag = 10;

/** Read location data from (valid!) XML node. */
static Location parseLocation(const pugi::xml_node &locationNode)
{
    // Position
    pugi::xml_node posNode = locationNode.child("position");
    Ogre::Real posX = posNode.attribute("x").as_double();
    Ogre::Real posY = posNode.attribute("y").as_double();
    Ogre::Real posZ = posNode.attribute("z").as_double();

    // Rotation
    pugi::xml_node rotNode = locationNode.child("rotation");
    Ogre::Real rotX = rotNode.attribute("qx").as_double();
    Ogre::Real rotY = rotNode.attribute("qy").as_double();
    Ogre::Real rotZ = rotNode.attribute("qz").as_double();
    Ogre::Real rotW = rotNode.attribute("qw").as_double();

    // Create location object
    Location location;
    location.position = Ogre::Vector3(posX, posY, posZ);
    location.rotation = Ogre::Quaternion(rotW, rotX, rotY, rotZ);

    return location;
}

Level::Level()
{
}

Level::~Level()
{
    delete m_trackBody;
    delete m_trackState;
    delete m_trackTriMesh;

    delete m_borderBody;
    delete m_borderState;
    delete m_borderTriMesh;

    for (int i = 0; i < m_accelerators.size(); i++) {
        delete m_accelerators[i].body;
        delete m_accelerators[i].state;
        delete m_accelerators[i].shape;
    }

    for (int i = 0; i < m_checkpoints.size(); i++) {
        delete m_checkpoints[i].body;
        delete m_checkpoints[i].state;
    }

    delete m_triggerShape;

    m_instance = NULL;
}

Level *Level::getSingleton()
{
    if (m_instance == NULL)
        m_instance = new Level;

    return m_instance;
}

void Level::create(std::string levelName)
{
    // Load level structure from XML file
    pugi::xml_document doc;
    pugi::xml_parse_result result = doc.load_file((levelDir + levelName + levelExt).c_str());

    // TODO: User friendlier error window
    if (!result) {
        std::cerr << "Level loading error" << std::endl;
        return;
    }
    
    // Basic XML nodes of the level
    pugi::xml_node xmlLevelNode = doc.child("level");
    pugi::xml_node xmlTrackNode = xmlLevelNode.child("track");
    pugi::xml_node xmlBorderNode = xmlTrackNode.child("border");
    pugi::xml_node xmlLightsNode = xmlLevelNode.child("lights");
    pugi::xml_node xmlVisualsNode = xmlLevelNode.child("visuals");
    pugi::xml_node xmlCamerasNode = xmlLevelNode.child("cameras");
    pugi::xml_node xmlObjectsNode = xmlLevelNode.child("objects");
    pugi::xml_node xmlDocksNode = xmlLevelNode.child("docks");
    pugi::xml_node xmlCheckpointsNode = xmlLevelNode.child("checkpoints");
    
    // Create track's node and attach it to scene
    m_trackNode = m_sceneManager->createSceneNode();
    m_sceneManager->getRootSceneNode()->addChild(m_trackNode);

    // Create track's entity using external mesh
    m_trackEntity = m_sceneManager->createEntity(xmlTrackNode.attribute("file").value());
    m_trackEntity->setCastShadows(false);
    m_trackNode->attachObject(m_trackEntity);

    // Create border's node and attach it to scene
    m_borderNode = m_sceneManager->createSceneNode();
    m_sceneManager->getRootSceneNode()->addChild(m_borderNode);

    // Create border's entity using external mesh
    m_borderEntity = m_sceneManager->createEntity(xmlBorderNode.attribute("file").value());
    m_borderEntity->setCastShadows(false);
    m_borderNode->attachObject(m_borderEntity);

    // The same way we load all other (decorative) objects
    for (pugi::xml_node vis = xmlVisualsNode.child("object"); vis; vis = vis.next_sibling()) {
        // Create object's node and attach it to scene
        Ogre::SceneNode *objNode = m_sceneManager->createSceneNode();
        m_sceneManager->getRootSceneNode()->addChild(objNode);

        // Create object's entity using external mesh
        Ogre::Entity *objEntity = m_sceneManager->createEntity(vis.attribute("file").value());
        objEntity->setCastShadows(false);
        objNode->attachObject(objEntity);
    }

    // Load physical objects
    unsigned int idAccel = acceleratorMinId;
    unsigned int idSlow = slowdownMinId;
    for (pugi::xml_node obj = xmlObjectsNode.child("object"); obj; obj = obj.next_sibling()) {
        // Parse location parameters for this object
        Location objLoc = parseLocation(obj);

        // Create object's node and attach it to scene
        Ogre::SceneNode *objNode = m_sceneManager->createSceneNode();
        objNode->setPosition(objLoc.position);
        objNode->setOrientation(objLoc.rotation);
        m_sceneManager->getRootSceneNode()->addChild(objNode);

        // Create object's entity using external mesh
        Ogre::Entity *objEntity = m_sceneManager->createEntity(obj.attribute("file").value());
        objEntity->setCastShadows(false);
        objNode->attachObject(objEntity);

        // Object trigger entity will be used later in physics initialization
        Ogre::Entity *objTrigger = m_sceneManager->createEntity(obj.attribute("trigger").value());

        // Create accelerator
        Accelerator accel;
        accel.position = objLoc.position;
        accel.rotation = objLoc.rotation;
        accel.node = objNode;
        accel.entity = objTrigger;

        // It can be either accelerator or slowdown object
        if (std::string(obj.attribute("type").value()) == "accelerator") {
            accel.id = idAccel++;
        } else {
            accel.id = idSlow++;
        }

        // Add it to the list
        m_accelerators.push_back(accel);
    }

    // Load starting positions and rotations
    for (pugi::xml_node dock = xmlDocksNode.child("dock"); dock; dock = dock.next_sibling()) {
        // Parse location parameters for this dock and add it to the list
        m_docks.push_back(parseLocation(dock));
    }

    // Checkpoint trigger entity will be used later in physics initialization
    m_triggerEntity = m_sceneManager->createEntity(xmlCheckpointsNode.attribute("trigger").value());

    // Load checkpoints
    unsigned int id = checkpointMinId;
    for (pugi::xml_node cp = xmlCheckpointsNode.child("checkpoint"); cp; cp = cp.next_sibling()) {
        // Parse location parameters for this checkpoint
        Location cpLoc = parseLocation(cp);

        // Scene node for the checkpoint
        Ogre::SceneNode *cpNode = m_sceneManager->createSceneNode();
        cpNode->setPosition(cpLoc.position);
        cpNode->setOrientation(cpLoc.rotation);

        // Add checkpoint to the list
        Checkpoint checkPoint;
        checkPoint.position = cpLoc.position;
        checkPoint.rotation = cpLoc.rotation;
        checkPoint.node = cpNode;
        checkPoint.id = id++;
        m_checkpoints.push_back(checkPoint);
    }

    // Load lights
    for (pugi::xml_node ln = xmlLightsNode.child("light"); ln; ln = ln.next_sibling()) {
        // Position
        pugi::xml_node pos = ln.child("position");
        Ogre::Real x = pos.attribute("x").as_double();
        Ogre::Real y = pos.attribute("y").as_double();
        Ogre::Real z = pos.attribute("z").as_double();

        // Color
        pugi::xml_node col = ln.child("color");
        Ogre::Real r = col.attribute("r").as_double();
        Ogre::Real g = col.attribute("g").as_double();
        Ogre::Real b = col.attribute("b").as_double();

        // Create the light
        Ogre::Light *light = m_sceneManager->createLight();
        light->setType(Ogre::Light::LT_POINT);
        light->setPosition(x, y, z);
        light->setDiffuseColour(r, g, b);
    }

    // Load camera positions
    for (pugi::xml_node cn = xmlCamerasNode.child("camera"); cn; cn = cn.next_sibling()) {
        // Position
        pugi::xml_node pos = cn.child("position");
        Ogre::Real x = pos.attribute("x").as_double();
        Ogre::Real y = pos.attribute("y").as_double();
        Ogre::Real z = pos.attribute("z").as_double();

        // Add camera position to the list
        m_cameras.push_back(Ogre::Vector3(x, y, z));
    }

    this->initPhysics();
}

Location Level::getDock(unsigned int index) const
{
    if (index >= m_docks.size()) {
        Location loc;
        return loc;
    }

    return m_docks[index];
}

Location Level::getCheckpoint(unsigned int index) const
{
    if (index >= m_checkpoints.size()) {
        Location loc;
        return loc;
    }

    return m_checkpoints[index];
}

unsigned int Level::getPrevCheckpoint(unsigned int cp) const
{
    if (cp == 0)
        return (m_checkpoints.size() - 1);
    else
        return (cp - 1);
}

unsigned int Level::getNextCheckpoint(unsigned int cp) const
{
    return ((cp + 1) % m_checkpoints.size());
}

unsigned int Level::getFirstCheckpoint() const
{
    return 0;
}

unsigned int Level::getHalfCheckpoint() const
{
    return (m_checkpoints.size() / 2);
}

void Level::initAccelerators()
{
    // Add all accelerator triggers to the physical simulation
    for (int i = 0; i < m_accelerators.size(); i++) {
        // Create collision shape
        BtOgre::StaticMeshToShapeConverter converter(m_accelerators[i].entity);
        m_accelerators[i].shape = converter.createBox();

        // Create bullet motion state and static rigid body
        BtOgre::RigidBodyState *state = new BtOgre::RigidBodyState(m_accelerators[i].node);
        btRigidBody *body = new btRigidBody(0, state, m_accelerators[i].shape);

        // We don't want the accelerator triggers to stop the ship, so no contact response
        body->setCollisionFlags(body->getCollisionFlags() | btCollisionObject::CF_NO_CONTACT_RESPONSE);

        // Tag the accelerator so we can identify it in the physical world
        body->setUserPointer(static_cast<unsigned int *>(&m_accelerators[i].id));

        // Add body to the simulation
        Physics::getSingleton()->getWorld()->addRigidBody(body);

        // Save pointers for deletion in destructor
        m_accelerators[i].state = state;
        m_accelerators[i].body = body;
    }
}

void Level::initCheckpoints()
{
    // We will share the same shape for every checkpoint
    BtOgre::StaticMeshToShapeConverter converter(m_triggerEntity);
    m_triggerShape = converter.createBox();

    // Add all checkpoints to the physical simulation
    for (int i = 0; i < m_checkpoints.size(); i++) {
        // Create bullet motion state and static rigid body
        BtOgre::RigidBodyState *state = new BtOgre::RigidBodyState(m_checkpoints[i].node);
        btRigidBody *body = new btRigidBody(0, state, m_triggerShape);

        // We don't want the checkpoints to stop the ship, so no contact response
        body->setCollisionFlags(body->getCollisionFlags() | btCollisionObject::CF_NO_CONTACT_RESPONSE);

        // Tag the checkpoint so we can identify it in the physical world
        body->setUserPointer(static_cast<unsigned int *>(&m_checkpoints[i].id));
        
        // Add body to the simulation
        Physics::getSingleton()->getWorld()->addRigidBody(body);

        // Save pointers for deletion in destructor
        m_checkpoints[i].state = state;
        m_checkpoints[i].body = body;
    }
}

void Level::initTrackPhysics(Ogre::Entity *entity, btRigidBody **body,
    BtOgre::TriangleMesh **triMesh, btDefaultMotionState **state, bool track)
{
    // Create track's triangle mesh and shape
    BtOgre::StaticMeshToShapeConverter converter(entity);
    *triMesh = converter.createTrimesh();

    // Create MotionState
    *state = new btDefaultMotionState(btTransform(btQuaternion(0, 0, 0, 1), btVector3(0, 0, 0)));

    // Create track's body
    *body = new btRigidBody(0, *state, (*triMesh)->getShape(), btVector3(0, 0, 0));
    (*body)->setFriction(0.0);
    (*body)->setRestitution(0.0);

    // Track is marked with special tag so we can later do proper ray tests
    if (track)
        (*body)->setUserPointer(static_cast<int*>(&trackTag));

    // Add it to the physics engine
    Physics::getSingleton()->getWorld()->addRigidBody(*body);
}

void Level::initPhysics()
{
    initTrackPhysics(m_trackEntity, &m_trackBody, &m_trackTriMesh, &m_trackState, true);
    initTrackPhysics(m_borderEntity, &m_borderBody, &m_borderTriMesh, &m_borderState, false);
    initAccelerators();
    initCheckpoints();
}
