/**
 *  @author Marian Ondrejicek (xondre02)
 *  @date   May 2012
 *
 *  Class that represents level of the game.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef LEVEL_H
#define LEVEL_H

#include "physics.h"
#include "levelobjects.h"

#include <OGRE/Ogre.h>

/** Level is a singleton that holds both visual and physical
  * parameters of the currently loaded race environment. */
class Level
{
protected:
    Level();
    ~Level();

public:
    /** Get pointer to the actual object. */
    static Level *getSingleton();

    /** Reset object to its initial state. */
    void destroy() { delete this; }

    void create(std::string levelName);
    void setSceneManager(Ogre::SceneManager *manager) { m_sceneManager = manager; }

    /** Return position where ship should start and how it should be rotated. */
    Location getDock(unsigned int index) const;

    /** Return location of given checkpoint. */
    Location getCheckpoint(unsigned int) const;

    /** Return ID of previous checkpoint with respect to looping. */
    unsigned int getPrevCheckpoint(unsigned int cp) const;

    /** Return ID of next checkpoint with respect to looping. */
    unsigned int getNextCheckpoint(unsigned int cp) const;

    /** Return ID of first checkpoint (it is the start/finish). */
    unsigned int getFirstCheckpoint() const;

    /** Return ID of checkpoint in the half way of track. */
    unsigned int getHalfCheckpoint() const;

    /** Return pointer to list of cameras' postions. */
    const std::vector<Ogre::Vector3> *getCameras() { return &m_cameras; }

protected:
    /* Create physical objects from accelerators and add them to the simulation. */
    void initAccelerators();

    /* Create physical objects from checkpoints and add them to the simulation. */
    void initCheckpoints();

    /* This method is used for initializing physics of track and border. */
    void initTrackPhysics(Ogre::Entity *entity, btRigidBody **body,
        BtOgre::TriangleMesh **triMesh, btDefaultMotionState **state, bool track);

    /** Initialize all physical objects and add them to simulation. */
    void initPhysics();

private:
    static Level *m_instance;           /**< The only instance of this class. */

    Ogre::SceneManager *m_sceneManager; /**< Manages all objects in the scene. */

    Ogre::SceneNode *m_trackNode;
    Ogre::Entity *m_trackEntity;
    btRigidBody *m_trackBody;
    BtOgre::TriangleMesh *m_trackTriMesh;
    btDefaultMotionState *m_trackState;

    Ogre::SceneNode *m_borderNode;
    Ogre::Entity *m_borderEntity;
    btRigidBody *m_borderBody;
    BtOgre::TriangleMesh *m_borderTriMesh;
    btDefaultMotionState *m_borderState;

    Ogre::Entity *m_triggerEntity;              /**< Entity of the checkpoint trigger. */
    btCollisionShape *m_triggerShape;           /**< Shape of the checkpoint trigger. */

    std::vector<Accelerator> m_accelerators;    /**< List of all accelerators. */
    std::vector<Location> m_docks;              /**< Where the ship should start. */
    std::vector<Checkpoint> m_checkpoints;      /**< Sorted list of all checkpoints. */
    std::vector<Ogre::Vector3> m_cameras;       /**< Positions of all cameras. */
};

#endif  // LEVEL_H
