/**
 *  @author Marian Ondrejicek (xondre02)
 *  @date   May 2012
 *
 *  Initial screen of the game.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "menustate.h"
#include "levelobjects.h"
#include "soundmanager.h"
#include "racetimer.h"
#include "config.h"

// How fast the ships should rotate
const double rotateSpeed = 0.0005;

const std::string leftShip = "blaster";
const std::string rightShip = "proton";

// States of the finite-state machine, which is used to determine which menu screen should be active
enum MenuScreenStates
{
    STATE_MAINMENU,
    STATE_SELECTNAME,
    STATE_SELECTNAME1,
    STATE_SELECTNAME2,
    STATE_SELECTSHIP,
    STATE_SELECTSHIP1,
    STATE_SELECTSHIP2,
    STATE_SCOREBOARD,
    STATE_OPTIONS,
    STATE_CREDITS,
};

MenuState::MenuState()
{
    m_ship1 = new Ship;
    m_ship2 = new Ship;
    m_quit = false;
}

MenuState::~MenuState()
{
    delete m_ship1;
    delete m_ship2;
}

void MenuState::enter()
{
    // Create scene manager for this state
    m_sceneManager = OgreFramework::getSingletonPtr()->
            m_root->createSceneManager(Ogre::ST_GENERIC, "MenuSceneManager");

    // Connect the scene manager with sound manager
    SoundManager::getSingleton()->attach(m_sceneManager);

    // Enable/disable music globally
    bool music = Config::getSingleton()->isMusicEnabled();
    SoundManager::getSingleton()->setMute(!music);

    // Get pointer to viewport
    Ogre::Viewport *viewport = OgreFramework::getSingletonPtr()->m_viewport;

    // Setup camera
    Ogre::Camera *cam = m_sceneManager->createCamera("MenuCamera");
    cam->setNearClipDistance(5);
    cam->setAspectRatio(Ogre::Real(viewport->getActualWidth())/
                           Ogre::Real(viewport->getActualHeight()));

    cam->setPosition(20, 10, 20);
    cam->lookAt(0, 5, 0);

    // Associate camera with viewport
    viewport->setCamera(cam);

    // Associate GUI with the scene
    Gui::getSingleton()->attach(m_sceneManager);
    Gui::getSingleton()->getPlatform()->getRenderManagerPtr()->setActiveViewport(0);

    // Show the mouse cursor (just in case it was hidden before)
    MyGUI::PointerManager::getInstance().setVisible(true);

    // Enter the initial menu screen state
    m_currentState = STATE_MAINMENU;

    // Get pointers to materials of each ship
    m_ship1Mat = static_cast<Ogre::MaterialPtr>
        (Ogre::MaterialManager::getSingleton().getByName("BlasterMaterial"));
    m_ship2Mat = static_cast<Ogre::MaterialPtr>
        (Ogre::MaterialManager::getSingleton().getByName("Proton_auv"));

    // Save original specular colors of each ship
    m_ship1Specular = m_ship1Mat->getTechnique(0)->getPass(0)->getSpecular();
    m_ship2Specular = m_ship2Mat->getTechnique(0)->getPass(0)->getSpecular();

    // Set specular color of each ship to white
    m_ship1Mat->getTechnique(0)->getPass(0)->setSpecular(1.0, 1.0, 1.0, 1.0);
    m_ship2Mat->getTechnique(0)->getPass(0)->setSpecular(1.0, 1.0, 1.0, 1.0);

    createScene();
    showGui();
}

void MenuState::exit()
{
    // Disconnect the GUI
    Gui::getSingleton()->detach();

    // Disconnect the sound manager
    SoundManager::getSingleton()->detach();
    
    // Destroy the scene manager
    if (m_sceneManager)
        OgreFramework::getSingletonPtr()->m_root->destroySceneManager(m_sceneManager);

    // Restore original specular colors of the ships
    m_ship1Mat->getTechnique(0)->getPass(0)->setSpecular(m_ship1Specular);
    m_ship2Mat->getTechnique(0)->getPass(0)->setSpecular(m_ship2Specular);
}

bool MenuState::keyPressed(const OIS::KeyEvent &arg)
{
    MyGUI::InputManager::getInstance().injectKeyPress(MyGUI::KeyCode::Enum(arg.key), arg.text);

    // Pressing escape will show the main menu
    if (arg.key == OIS::KC_ESCAPE) {
        if (m_currentState != STATE_MAINMENU) {
            m_currentState = STATE_MAINMENU;
            showGui();
        }
    }

    // Pressing enter when typing nickname equals clicking on OK button
    if (arg.key == OIS::KC_RETURN) {
        switch (m_currentState)
        {
        case STATE_SELECTNAME:
        case STATE_SELECTNAME1:
        case STATE_SELECTNAME2:
            guiOkPressed(NULL);
            break;
        }
    }

    // Forward events to the framework
    OgreFramework::getSingletonPtr()->keyPressed(arg);
    return true;
}

bool MenuState::keyReleased(const OIS::KeyEvent &arg)
{
    MyGUI::InputManager::getInstance().injectKeyRelease(MyGUI::KeyCode::Enum(arg.key));

    // Forward events to the framework
    OgreFramework::getSingletonPtr()->keyReleased(arg);
    return true;
}

bool MenuState::mouseMoved(const OIS::MouseEvent &arg)
{
    MyGUI::InputManager::getInstance().injectMouseMove(arg.state.X.abs, arg.state.Y.abs, arg.state.Z.abs);
    return true;
}

bool MenuState::mousePressed(const OIS::MouseEvent &arg, OIS::MouseButtonID id)
{
    MyGUI::InputManager::getInstance().injectMousePress(arg.state.X.abs, arg.state.Y.abs, MyGUI::MouseButton::Enum(id));
    return true;
}

bool MenuState::mouseReleased(const OIS::MouseEvent &arg, OIS::MouseButtonID id)
{
    MyGUI::InputManager::getInstance().injectMouseRelease(arg.state.X.abs, arg.state.Y.abs, MyGUI::MouseButton::Enum(id));
    return true;
}

void MenuState::guiSinglePressed(MyGUI::Widget *)
{
    Config::getSingleton()->setSplitScreen(false);
    //changeAppState(findByName("GameState"));
    m_currentState = STATE_SELECTNAME;
    showGui();
}

void MenuState::guiSplitPressed(MyGUI::Widget *)
{
    Config::getSingleton()->setSplitScreen(true);
    //changeAppState(findByName("GameState"));
    m_currentState = STATE_SELECTNAME1;
    showGui();
}

void MenuState::guiScorePressed(MyGUI::Widget *)
{
    m_currentState = STATE_SCOREBOARD;
    showGui();
}

void MenuState::guiOptionsPressed(MyGUI::Widget *)
{
    m_currentState = STATE_OPTIONS;
    showGui();
}

void MenuState::guiCreditsPressed(MyGUI::Widget *)
{
    m_currentState = STATE_CREDITS;
    showGui();
}

void MenuState::guiExitPressed(MyGUI::Widget *)
{
    Config::getSingleton()->save();
    m_quit = true;
}

void MenuState::guiLeftShipPressed(MyGUI::Widget *)
{
    // We need to distinguish splitscreen and singleplayer Ship buttons
    switch (m_currentState) {
        case STATE_SELECTSHIP:
            Config::getSingleton()->getPlayer(PLAYER_SINGLE)->setShipName(leftShip);
            changeAppState(findByName("GameState"));
            return;
        case STATE_SELECTSHIP1:
            Config::getSingleton()->getPlayer(PLAYER_SPLIT1)->setShipName(leftShip);
            m_currentState = STATE_SELECTNAME2;
            showGui();
            break;
        case STATE_SELECTSHIP2:
            Config::getSingleton()->getPlayer(PLAYER_SPLIT2)->setShipName(leftShip);
            changeAppState(findByName("GameState"));
            return;
    }
}

void MenuState::guiRightShipPressed(MyGUI::Widget *)
{
    // We need to distinguish splitscreen and singleplayer Ship buttons
    switch (m_currentState) {
        case STATE_SELECTSHIP:
            Config::getSingleton()->getPlayer(PLAYER_SINGLE)->setShipName(rightShip);
            changeAppState(findByName("GameState"));
            return;
        case STATE_SELECTSHIP1:
            Config::getSingleton()->getPlayer(PLAYER_SPLIT1)->setShipName(rightShip);
            m_currentState = STATE_SELECTNAME2;
            showGui();
            break;
        case STATE_SELECTSHIP2:
            Config::getSingleton()->getPlayer(PLAYER_SPLIT2)->setShipName(rightShip);
            changeAppState(findByName("GameState"));
            return;
    }
}

void MenuState::guiSavePressed(MyGUI::Widget *)
{
    // Find out what options has user checked
    MyGUI::Gui *myGui = Gui::getSingleton()->getGui();

    MyGUI::ButtonPtr blurCheckBox = myGui->findWidget<MyGUI::Button>("BlurCheckBox");
    MyGUI::ButtonPtr glowCheckBox = myGui->findWidget<MyGUI::Button>("GlowCheckBox");
    MyGUI::ButtonPtr musicCheckBox = myGui->findWidget<MyGUI::Button>("MusicCheckBox");

    bool blur = blurCheckBox->getStateSelected();
    bool glow = glowCheckBox->getStateSelected();
    bool music = musicCheckBox->getStateSelected();

    // Apply new settings
    Config::getSingleton()->setBlurEnabled(blur);
    Config::getSingleton()->setGlowEnabled(glow);
    Config::getSingleton()->setMusicEnabled(music);

    // Music is present in menu, so we should enable/disable it immediately
    SoundManager::getSingleton()->setMute(!music);

    // Go back to menu
    m_currentState = STATE_MAINMENU;
    showGui();
}

void MenuState::guiOkPressed(MyGUI::Widget *)
{
    // Find out what nickname has user typed in
    MyGUI::Gui *myGui = Gui::getSingleton()->getGui();
    MyGUI::EditBox *nameEditBox = myGui->findWidget<MyGUI::EditBox>("NameEditBox");
    std::string nick = nameEditBox->getCaption();

    // We need to distinguish splitscreen and singleplayer OK buttons
    switch (m_currentState) {
        case STATE_SELECTNAME:
            Config::getSingleton()->getPlayer(PLAYER_SINGLE)->setNick(nick);
            m_currentState = STATE_SELECTSHIP;
            break;
        case STATE_SELECTNAME1:
            Config::getSingleton()->getPlayer(PLAYER_SPLIT1)->setNick(nick);
            m_currentState = STATE_SELECTSHIP1;
            break;
        case STATE_SELECTNAME2:
            Config::getSingleton()->getPlayer(PLAYER_SPLIT2)->setNick(nick);
            m_currentState = STATE_SELECTSHIP2;
            break;
    }
    
    showGui();
}

void MenuState::guiCancelPressed(MyGUI::Widget *)
{
    m_currentState = STATE_MAINMENU;
    showGui();
}

void MenuState::guiCheckBoxPressed(MyGUI::Widget *widget)
{
    // Toggle the checkbox
    MyGUI::ButtonPtr checkBox = widget->castType<MyGUI::Button>();
    checkBox->setStateSelected(!checkBox->getStateSelected());
}

void MenuState::update(double timeSinceLastFrame)
{
    if (m_quit)
        shutdown();

    m_ship1->getNode()->yaw(Ogre::Radian(timeSinceLastFrame * rotateSpeed));
    m_ship2->getNode()->yaw(Ogre::Radian(timeSinceLastFrame * rotateSpeed));
}

void MenuState::createScene()
{
    // Setup sounds
    SoundManager::getSingleton()->createMenuSounds();
    SoundManager::getSingleton()->play("MenuSong");

    m_sceneManager->setShadowTechnique(Ogre::SHADOWTYPE_STENCIL_MODULATIVE);
    m_sceneManager->setShadowColour(Ogre::ColourValue(0.5, 0.5, 0.5));

    Ogre::SceneNode *lightNode = m_sceneManager->createSceneNode();
    m_sceneManager->getRootSceneNode()->addChild(lightNode);
    Ogre::Light *light = m_sceneManager->createLight("light");
    light->setType(Ogre::Light::LT_POINT);
    light->setPosition(0, 40.0, 0);
    light->setDiffuseColour(1.0f, 1.0f, 1.0f);

    Location loc1;
    loc1.position = Ogre::Vector3(-5.33, 0.9, 8.36);
    m_ship1->setSceneManager(m_sceneManager);
    m_ship1->create("blaster", loc1);

    Location loc2;
    loc2.position = Ogre::Vector3(8.34, 0.9, -5.35);
    m_ship2->setSceneManager(m_sceneManager);
    m_ship2->create("proton", loc2);

    // Create dome node and attach it to scene
    Ogre::SceneNode *domeNode = m_sceneManager->createSceneNode();
    m_sceneManager->getRootSceneNode()->addChild(domeNode);

    // Create dome entity using external mesh
    Ogre::Entity *domeEntity = m_sceneManager->createEntity("GeoDome.mesh");
    domeEntity->setCastShadows(false);
    domeNode->attachObject(domeEntity);

    // Create floor node and attach it to scene
    Ogre::SceneNode *floorNode = m_sceneManager->createSceneNode();
    m_sceneManager->getRootSceneNode()->addChild(floorNode);

    // Create floor entity using external mesh
    Ogre::Entity *floorEntity = m_sceneManager->createEntity("BottomCube.mesh");
    floorEntity->setCastShadows(false);
    floorNode->attachObject(floorEntity);

    // Create 1st stand node and attach it to scene
    Ogre::SceneNode *stand1Node = m_sceneManager->createSceneNode();
    m_sceneManager->getRootSceneNode()->addChild(stand1Node);

    // Create 1st stand entity using external mesh
    Ogre::Entity *stand1Entity = m_sceneManager->createEntity("cylinder1.mesh");
    stand1Entity->setCastShadows(false);
    stand1Node->attachObject(stand1Entity);

    // Create 2nd stand node and attach it to scene
    Ogre::SceneNode *stand2Node = m_sceneManager->createSceneNode();
    m_sceneManager->getRootSceneNode()->addChild(stand2Node);

    // Create 2nd stand entity using external mesh
    Ogre::Entity *stand2Entity = m_sceneManager->createEntity("cylinder2.mesh");
    stand2Entity->setCastShadows(false);
    stand2Node->attachObject(stand2Entity);
}

void MenuState::showGui()
{
    MyGUI::Gui *myGui = Gui::getSingleton()->getGui();

    // Show different GUI layouts according to current state
    switch (m_currentState) {
        case STATE_SELECTNAME:
        case STATE_SELECTNAME1:
        case STATE_SELECTNAME2:
            {
                Gui::getSingleton()->load("MenuStateName.layout");
                MyGUI::ButtonPtr okButton = myGui->findWidget<MyGUI::Button>("OkButton");
                okButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiOkPressed);
                MyGUI::ButtonPtr cancelButton = myGui->findWidget<MyGUI::Button>("CancelButton");
                cancelButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiCancelPressed);

                // Just change controls image to WASD
                if (m_currentState == STATE_SELECTNAME2) {
                    MyGUI::ImageBox *controlsImageBox = myGui->findWidget<MyGUI::ImageBox>("ControlsImageBox");
                    controlsImageBox->setImageTexture("wasd.png");
                }

                // Set last used nickname for this player, so he doesn't need to type
                std::string nick;
                if (m_currentState == STATE_SELECTNAME1)
                    nick = Config::getSingleton()->getPlayer(PLAYER_SPLIT1)->getNick();
                else if (m_currentState == STATE_SELECTNAME2)
                    nick = Config::getSingleton()->getPlayer(PLAYER_SPLIT2)->getNick();
                else
                    nick = Config::getSingleton()->getPlayer(PLAYER_SINGLE)->getNick();
                myGui->findWidget<MyGUI::EditBox>("NameEditBox")->setCaption(nick);
            }
            break;

        case STATE_SELECTSHIP:
        case STATE_SELECTSHIP1:
        case STATE_SELECTSHIP2:
            {
                Gui::getSingleton()->load("MenuStateShip.layout");
                MyGUI::ButtonPtr leftShipButton = myGui->findWidget<MyGUI::Button>("LeftShipButton");
                leftShipButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiLeftShipPressed);
                MyGUI::ButtonPtr rightShipButton = myGui->findWidget<MyGUI::Button>("RightShipButton");
                rightShipButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiRightShipPressed);
                MyGUI::ButtonPtr cancelButton = myGui->findWidget<MyGUI::Button>("CancelButton");
                cancelButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiCancelPressed);

                // Just change controls image to WASD
                if (m_currentState == STATE_SELECTSHIP2) {
                    MyGUI::ImageBox *controlsImageBox = myGui->findWidget<MyGUI::ImageBox>("ControlsImageBox");
                    controlsImageBox->setImageTexture("wasd.png");
                }
            }
            break;

        case STATE_SCOREBOARD:
            {
                Gui::getSingleton()->load("MenuStateScoreboard.layout");
                MyGUI::ButtonPtr okButton = myGui->findWidget<MyGUI::Button>("OkButton");
                okButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiCancelPressed);

                // Fill the single player scoreboard
                const std::vector<Score> &scoresSingle = Config::getSingleton()->getScoreboardSingle();
                for (int i = 1; i <= scoresSingle.size() ; i++) {
                    const std::string &nick = scoresSingle[i-1].nick;
                    const std::string &time = RaceTimer::timeToString(scoresSingle[i-1].time, true);
                    std::string row = Ogre::StringConverter::toString(i);
                    myGui->findWidget<MyGUI::TextBox>("SingleScoreName" + row)->setCaption(row + ". " + nick);
                    myGui->findWidget<MyGUI::TextBox>("SingleScoreTime" + row)->setCaption(time);
                }

                // Fill the split screen scoreboard
                const std::vector<Score> &scoresSplit= Config::getSingleton()->getScoreboardSplit();
                for (int i = 1; i <= scoresSplit.size() ; i++) {
                    const std::string &nick = scoresSplit[i-1].nick;
                    const std::string &time = RaceTimer::timeToString(scoresSplit[i-1].time, true);
                    std::string row = Ogre::StringConverter::toString(i);
                    myGui->findWidget<MyGUI::TextBox>("SplitScoreName" + row)->setCaption(row + ". " + nick);
                    myGui->findWidget<MyGUI::TextBox>("SplitScoreTime" + row)->setCaption(time);
                }
            }
            break;

        case STATE_OPTIONS:
            {
                Gui::getSingleton()->load("MenuStateOptions.layout");
                MyGUI::ButtonPtr saveButton = myGui->findWidget<MyGUI::Button>("SaveButton");
                saveButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiSavePressed);
                MyGUI::ButtonPtr cancelButton = myGui->findWidget<MyGUI::Button>("CancelButton");
                cancelButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiCancelPressed);

                // Checkboxes
                MyGUI::ButtonPtr blurCheckBox = myGui->findWidget<MyGUI::Button>("BlurCheckBox");
                blurCheckBox->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiCheckBoxPressed);
                blurCheckBox->setStateSelected(Config::getSingleton()->isBlurEnabled());
                MyGUI::ButtonPtr glowCheckBox = myGui->findWidget<MyGUI::Button>("GlowCheckBox");
                glowCheckBox->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiCheckBoxPressed);
                glowCheckBox->setStateSelected(Config::getSingleton()->isGlowEnabled());
                MyGUI::ButtonPtr musicCheckBox = myGui->findWidget<MyGUI::Button>("MusicCheckBox");
                musicCheckBox->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiCheckBoxPressed);
                musicCheckBox->setStateSelected(Config::getSingleton()->isMusicEnabled());
            }
            break;

        case STATE_CREDITS:
            {
                Gui::getSingleton()->load("MenuStateCredits.layout");
                MyGUI::ButtonPtr okButton = myGui->findWidget<MyGUI::Button>("OkButton");
                okButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiCancelPressed);
            }
            break;

        case STATE_MAINMENU:
        default:
            {
                Gui::getSingleton()->load("MenuState.layout");
                MyGUI::ButtonPtr singleButton = myGui->findWidget<MyGUI::Button>("SinglePlayerButton");
                singleButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiSinglePressed);
                MyGUI::ButtonPtr splitButton = myGui->findWidget<MyGUI::Button>("SplitScreenButton");
                splitButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiSplitPressed);
                MyGUI::ButtonPtr scoreButton = myGui->findWidget<MyGUI::Button>("ScoreBoardButton");
                scoreButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiScorePressed);
                MyGUI::ButtonPtr optionsButton = myGui->findWidget<MyGUI::Button>("OptionsButton");
                optionsButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiOptionsPressed);
                MyGUI::ButtonPtr creditsButton = myGui->findWidget<MyGUI::Button>("CreditsButton");
                creditsButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiCreditsPressed);
                MyGUI::ButtonPtr exitButton = myGui->findWidget<MyGUI::Button>("ExitButton");
                exitButton->eventMouseButtonClick += MyGUI::newDelegate(this, &MenuState::guiExitPressed);
            }
            break;
    }
}
