/**
 *  @author Marian Ondrejicek (xondre02)
 *  @date   May 2012
 *
 *  This class connects the game with the Bullet physics engine.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "physics.h"
#include "trackraycallback.h"

#include <bullet/BulletCollision/CollisionDispatch/btInternalEdgeUtility.h>

// At the beginning, we have no instance
Physics *Physics::m_instance = NULL;

// Gravitation constant
const double g = -9.8;

// Distance from ship to the ground
const double trackDist = 2.0;

// Ray length for ray casting
const double rayLength = 4.0;

// Milliseconds of extra boost and slowdown
const float accelerationMs = 2000;
const float slowdownMs = 2000;

void preTickCallback(btDynamicsWorld *world, btScalar timeStep)
{
    Physics *physics = static_cast<Physics *>(world->getWorldUserInfo());
    physics->processCallback(timeStep);
}

static bool CustomMaterialCombinerCallback(btManifoldPoint& cp,	const btCollisionObject* colObj0,
    int partId0, int index0, const btCollisionObject* colObj1, int partId1, int index1)
{
    if (static_cast<Player *>(colObj1->getUserPointer())) {
        Player *player = static_cast<Player *>(colObj1->getUserPointer());
        
        // Did we hit some tagged object?
        if (static_cast<unsigned int *>(colObj0->getUserPointer())) {
            // Get ID of that object
            unsigned int id = *static_cast<unsigned int *>(colObj0->getUserPointer());

            // Check the ID ranges to see what type of object ship has hit
            if (id >= acceleratorMinId && id <= acceleratorMaxId) {
                // Ship has hit an accelerator
                player->getShip()->addAcceleration(accelerationMs);
            } else if (id >= slowdownMinId && id <= slowdownMaxId) {
                // Ship has hit a slowdown
                player->getShip()->addSlowdown(slowdownMs);
            } else if (id >= checkpointMinId && id <= checkpointMaxId) {
                // Ship has hit a checkpoint
                player->setCurrentCheckpoint(id);
            }
        }
    }

    // This ensures that the ship's movement is smooth while driving near borders
    btAdjustInternalEdgeContacts(cp, colObj1, colObj0, partId1, index1);
    
    cp.m_combinedFriction = 0;

    // Return false because we did not compute the friction
    return false;
}

extern ContactAddedCallback gContactAddedCallback;

Physics::Physics()
{
    // Bullet initialization
    m_broadphase = new btAxisSweep3(btVector3(-10000,-10000,-10000), btVector3(10000,10000,10000), 1024);
    m_collisionConfig = new btDefaultCollisionConfiguration();
    m_dispatcher = new btCollisionDispatcher(m_collisionConfig);
    m_solver = new btSequentialImpulseConstraintSolver();

    m_world = new btDiscreteDynamicsWorld(m_dispatcher, m_broadphase, m_solver, m_collisionConfig);
    m_world->setGravity(btVector3(0, g, 0));

    // Just some drawing stuff for debugging purposes
    m_dbgdraw = new BtOgre::DebugDrawer(Ogre::Root::getSingletonPtr()->
        getSceneManager("GameSceneManager")->getRootSceneNode(), m_world);
    m_world->setDebugDrawer(m_dbgdraw);
    m_debug = false;
    m_dbgdraw->setDebugMode(m_debug);

    // Register callback function to be called before every simulation step
    m_world->setInternalTickCallback(preTickCallback, static_cast<void *>(this), true);

    // Use our custom material combiner function
    gContactAddedCallback = CustomMaterialCombinerCallback;
}

Physics::~Physics()
{
    delete m_dbgdraw;
    delete m_world;
    delete m_solver;
    delete m_dispatcher;
    delete m_collisionConfig;
    delete m_broadphase;

    m_instance = NULL;
}

Physics *Physics::getSingleton()
{
    if (m_instance == NULL)
        m_instance = new Physics;

    return m_instance;
}

void Physics::addShip(Ship *ship, Player *player)
{
    btRigidBody *body = ship->getBody();

    // Enable custom material callback
    body->setCollisionFlags(body->getCollisionFlags() | btCollisionObject::CF_CUSTOM_MATERIAL_CALLBACK);

    // Tag the body with player's pointer, so we can access it later when detecting collisions
    body->setUserPointer(static_cast<Player *>(player));

    m_world->addRigidBody(body);
    m_ships.push_back(ship);
}

void Physics::toggleDebug()
{
    m_debug = !m_debug;
    m_dbgdraw->setDebugMode(m_debug);
}

bool Physics::rayCast(const btVector3 &start, const btVector3 &end, btVector3 &hit)
{
    // Ray callback function
    TrackRayCallback ray(start, end);

    // Do the ray casting
    m_world->rayTest(start, end, ray);

    // Check if we have hit the track
    if (ray.hasHit()) {
        hit = ray.m_hitPointWorld;
        return true;
    }

    return false;
}

void Physics::processCallback(btScalar timeStep)
{
    // Iterate over all ships and apply appropriate forces
    for (std::vector<Ship*>::iterator it = m_ships.begin(); it != m_ships.end(); it++) {

        // Just to make the following code more readable
        Ship *ship = *it;

        btRigidBody *body = ship->getBody();
        const btMatrix3x3 &rotation = body->getWorldTransform().getBasis();
        const btAlignedObjectArray<btVector3> &springs = ship->getSprings();
        const btAlignedObjectArray<Ogre::SceneNode*> &markers = ship->getMarkers();

        body->clearForces();
        body->setAngularVelocity(btVector3(0, 0, 0));

        // Current speed of the ship
        double speed = body->getLinearVelocity().length();

        if (ship->isAccelerating() && speed < ship->getMaxSpeed()) {
            // Make the ship accelerate
            double power = ship->getAcceleration() * ship->getMass() * timeStep;
            btVector3 force = rotation * btVector3(0, 0, power);
            body->applyCentralForce(force);
        } else if (ship->isSlowingDown()) {
            // Make the ship slow down
            btVector3 force = -body->getLinearVelocity() * ship->getMass() * ship->getBreakPower();
            body->applyCentralForce(force);
        }
        else {
            // Make the ship slowly stop
            btVector3 force = -body->getLinearVelocity() * ship->getMass();
            body->applyCentralForce(force);
        }

        if (ship->isSteeringLeft() || ship->isSteeringRight()) {
            // Left or right direction?
            int dir = ship->isSteeringLeft() ? 1 : -1;
            // Rotate the ship
            double power = ship->getSteerPower() * ship->getMass() * timeStep;
            btVector3 torque = rotation * btVector3(0, dir*power, 0);
            body->applyTorque(torque);
            // Remember magnitude of current velocity
            btScalar magnitude = body->getLinearVelocity().length();
            // Stop the ship to make calculations easier
            body->setLinearVelocity(btVector3(0, 0, 0));
            // Move the ship in new direction but with the same speed as before
            body->applyCentralImpulse(rotation * btVector3(0, 0, magnitude*ship->getMass()));
        }

        // Global gravity of the world
        btVector3 worldGravity = btVector3(0, g * ship->getMass() * timeStep, 0);
        // Gravity of the track (to make the ship sit on the ground)
        btVector3 gravity = rotation * worldGravity;

        // How many springs are too high above the track (they are in space)
        int springsInSpace = 0;

        // Do the ray cast test for all four corners (springs)
        for (int i = 0; i < springs.size(); ++i) {
            btVector3 start = body->getCenterOfMassPosition() + (rotation * springs[i]);
            btVector3 endRelative(springs[i].x(), springs[i].y()-rayLength, springs[i].z());
            btVector3 end = body->getCenterOfMassPosition() + (rotation * endRelative);
            btVector3 hit;

            // Destination point, where the force will be applied
            btVector3 dest = rotation * springs[i];
            
            markers[i]->setVisible(false);

            // Do the ray casting
            if (rayCast(start, end, hit)) {
                btScalar dist = start.distance(hit);

                markers[i]->setVisible(m_debug);
                markers[i]->setPosition(BtOgre::Convert::toOgre(hit));
            
                // Absolute distance from the desired height
                float absDist = fabs(dist-trackDist);

                if (absDist > 2.0) {
                    // We are too high, just apply track's gravity
                    body->applyImpulse(gravity, dest);
                }
                else if (absDist > 0.1) {
                    // We should go a little higher/lower so we come closer to the desired height
                    body->applyForce(gravity*(dist-trackDist)*1000, dest);
                } else {
                    // This code is the same as code for controlling the ship.
                    // But this time it is used to avoid occasional trembling.

                    // Remember magnitude of current velocity
                    btScalar magnitude = body->getLinearVelocity().length();
                    // Stop the ship to make calculations easier
                    body->setLinearVelocity(btVector3(0, 0, 0));
                    // Move the ship in new direction but with the same speed as before
                    body->applyCentralImpulse(rotation * btVector3(0, 0, magnitude*ship->getMass()));
                }
            } else {
                // There is no ground under the ship, we are probably in space
                body->applyImpulse(worldGravity, dest);
                springsInSpace++;
            }
        }

        // If at least two springs are in space, we consider the whole ship to be in space
        if (springsInSpace >= 2) {
            ship->setInSpace(true);
        } else {
            ship->setInSpace(false);
        }
    }
}
