/**
 *  @author Marian Ondrejicek (xondre02)
 *  @date   May 2012
 *
 *  Class that represents player in the game.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef PLAYER_H
#define PLAYER_H

#include "ship.h"

#include <string>

/** Player's state in the race. */
enum PlayerState
{
    PLAYER_STATE_WAITING,
    PLAYER_STATE_RACING,
    PLAYER_STATE_FINISHED,
};

/** Player settings and state information. Each player should start
  * before first checkpoint so the computations are correct. */
class Player
{
public:
    Player();
    ~Player();

    /** Reset the variables to defaults. */
    void init();

    /** Set pointer to player's ship. */
    void setShip(Ship *ship) { m_ship = ship; }

    /** Return pointer to player's ship. */
    Ship *getShip() const { return m_ship; }

    /** Select ship for this player. Parameter must be a name of existing ship! */
    void setShipName(const std::string &shipName) { m_shipName = shipName; }

    /** Return name of the player's ship. */
    std::string getShipName() const { return m_shipName; }

    /** Set player's nickname. */
    void setNick(const std::string &nick) { m_nick = nick; }

    /** Return player's nickname. */
    std::string getNick() const { return m_nick; }

    /** Set whether we are going wrong way or not. This should be done in GameState. */
    void setWrongWay(bool value) { m_wrongWay = value; }

    /** Are we going wrong way? */
    bool isWrongWay() const { return m_wrongWay; }

    /** Get ID of last checkpoint we have passed. */
    unsigned int getCurrentCheckpoint() const;

    /** Set ID of last checkpoint we have passed. This should be done by Physics class. */
    void setCurrentCheckpoint(unsigned int cp);

    /** Get ID of checkpoint we expect to pass as next. */
    unsigned int getNextCheckpoint() const { return m_nextCheckpoint; }

    /** Set ID of checkpoint we expect to pass as next. This should be done in GameState. */
    void setNextCheckpoint(unsigned int cp) { m_nextCheckpoint = cp; }

    /** Set ID of first checkpoint (it is the start/finish). */
    void setFirstCheckpoint(unsigned int cp) { m_firstCheckpoint = cp; }

    /** Set ID of checkpoint in the half way of track (to check laps). */
    void setHalfCheckpoint(unsigned int cp) { m_halfCheckpoint = cp; }

    /** Get number of laps the player has traveled, including the currently unfinished. */
    unsigned int getLaps() const;

    /** Set number of half laps. */
    void setLaps(unsigned int laps);

    /** Get total race time. */
    unsigned int getTotalTime() const { return m_totalTime; }

    /** Set total time after finished race. */
    void setTotalTime(unsigned long time) { m_totalTime = time; }

    /** Get player's state in the race. */
    int getState() const { return m_state; }

    /** Set player's state in the race. */
    void setState(int state) { m_state = state; }

private:
    Ship *m_ship;               /**< Pointer to player's ship. */

    std::string m_nick;         /**< Player's nickname. */
    std::string m_shipName;     /**< Name of player's ship. */
    
    bool m_wrongWay;
    unsigned int m_currentCheckpoint;
    unsigned int m_nextCheckpoint;

    unsigned int m_firstCheckpoint;
    unsigned int m_halfCheckpoint;
    unsigned int m_lastHalf;    /**< ID of last checkpoint (first or half only). */
    unsigned int m_laps;        /**< Number of half laps the player has traveled. */

    unsigned int m_totalTime;   /**< Total race time in ms. */

    int m_state;                /**< Player's state in the race. */
};

#endif  // PLAYER_H
