/**
 *  @author Marian Ondrejicek (xondre02)
 *  @date   May 2012
 *
 *  Class that represents ship in the game.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "ship.h"

#include "pugixml/pugixml.hpp"

// Path to directory with ship files
#ifdef GAME_SHIPS_DIR
const std::string shipDir = GAME_SHIPS_DIR;
#else
const std::string shipDir = "../media/ships/";
#endif

// Extension of ship file
const std::string shipExt = ".ship";

// How long (ms) does it take until the ship steers at full power
const float steerPowerMax = 250;

// Maximal speed and acceleration are affected by these
// factors if extra boost/slowdown is applied.
const float accelerationFactor = 5;
const float slowdownFactor = 5;
const float maxSpeedFactor = 2;

// How long (milliseconds) can ship be in space before reseting back on track
const int inSpaceMax = 1800;

Ship::Ship()
{
    m_accelerate = false;
    m_slowDown = false;
    m_steerLeft = false;
    m_steerRight = false;

    m_steerPowerMs = 0;
    m_accelerationMs = 0;
    m_slowdownMs = 0;

    m_inSpace = false;
    m_inSpaceMs = 0;

    m_shipBody = NULL;
    m_shipState = NULL;
    m_shipShape = NULL;

    m_dataLoaded = false;
    m_physicsLoaded = false;
}

Ship::~Ship()
{
    if (m_shipBody)
        delete m_shipBody;
    if (m_shipState)
        delete m_shipState;
    if (m_shipShape)
        delete m_shipShape;
}

void Ship::create(const std::string &ship, const Location &loc)
{
    // Load ship structure from XML file
    pugi::xml_document doc;
    pugi::xml_parse_result result = doc.load_file((shipDir + ship + shipExt).c_str());

    // TODO: User friendlier error window
    if (!result) {
        std::cerr << "Ship loading error" << std::endl;
        return;
    }

    // Basic XML nodes of the ship
    pugi::xml_node xmlShipNode = doc.child("ship");
    pugi::xml_node xmlEngineNode = xmlShipNode.child("engine");

    // Create ship's node and attach it to scene.
    m_shipNode = m_sceneManager->createSceneNode();
    m_sceneManager->getRootSceneNode()->addChild(m_shipNode);
    m_shipNode->setPosition(loc.position);
    m_shipNode->setOrientation(loc.rotation);

    // Create ship's entity using external mesh
    m_shipEntity = m_sceneManager->createEntity(xmlShipNode.attribute("file").value());
    m_shipEntity->setCastShadows(true);
    m_shipNode->attachObject(m_shipEntity);

    // Load engine position
    Ogre::Real engineX = xmlEngineNode.attribute("x").as_double();
    Ogre::Real engineY = xmlEngineNode.attribute("y").as_double();
    Ogre::Real engineZ = xmlEngineNode.attribute("z").as_double();

    // Create node at engine's position
    Ogre::SceneNode *engineNode = m_sceneManager->createSceneNode();
    m_shipNode->addChild(engineNode);
    engineNode->setPosition(engineX, engineY, engineZ);

    // Generate unique name for particle system
    static int uniqueId = 0;
    std::stringstream sstream;
    sstream << uniqueId++;
    std::string partSystemName = "EngineFire" + sstream.str();

    // Create particle effect and attach it to engine
    m_partSystem = m_sceneManager->createParticleSystem(partSystemName, "Fire");
    engineNode->attachObject(m_partSystem);

    // Load some physical attributes as doubles
    m_mass = strtod(xmlShipNode.child("mass").child_value(), 0);
    m_acceleration = strtod(xmlShipNode.child("acceleration").child_value(), 0);
    m_maxSpeed = strtod(xmlShipNode.child("maxspeed").child_value(), 0);
    m_steerPower = strtod(xmlShipNode.child("steerpower").child_value(), 0);
    m_breakPower = strtod(xmlShipNode.child("breakpower").child_value(), 0);

    m_dataLoaded = true;
}

void Ship::attachCamera(Ogre::Camera *cam)
{
    if (!m_dataLoaded)
        return;
    
    m_shipNode->attachObject(cam);
}

void Ship::update(double timeSinceLastFrame, bool paused)
{
    if (!m_physicsLoaded)
        return;

    // Update velocity of the particles so they don't stay too behind when driving too fast
    const Ogre::Real initialVel = 10;
    Ogre::Real currentSpeed = m_shipBody->getLinearVelocity().length();
    if (paused)
        currentSpeed *= -1;
    Ogre::Real vel = initialVel - currentSpeed/2.0;
    m_partSystem->getEmitter(0)->setParticleVelocity(vel);

    if (!paused) {
        // Increase how long has the ship been steering
        if (m_steerLeft || m_steerRight) {
            m_steerPowerMs += timeSinceLastFrame;
            if (m_steerPowerMs > steerPowerMax)
                m_steerPowerMs = steerPowerMax;
        }

        // Decrease acceleration/slowdown milliseconds
        if (m_accelerationMs > 0)
            m_accelerationMs -= timeSinceLastFrame;
        if (m_slowdownMs > 0)
            m_slowdownMs -= timeSinceLastFrame;

        // Increase how long the ship is in space
        if (m_inSpace)
            m_inSpaceMs += timeSinceLastFrame;
    }
}

void Ship::steerLeft(bool val)
{
    m_steerLeft = val;

    if (!m_steerRight)
        m_steerPowerMs = 0;
}

void Ship::steerRight(bool val)
{
    m_steerRight = val;

    if (!m_steerRight)
        m_steerPowerMs = 0;
}

float Ship::getAcceleration() const
{
    if (m_accelerationMs > 0)
        return m_acceleration * accelerationFactor;

    if (m_slowdownMs > 0)
        return m_acceleration / slowdownFactor;

    return m_acceleration;
}

float Ship::getMaxSpeed() const
{
    if (m_accelerationMs > 0)
        return m_maxSpeed * maxSpeedFactor;

    if (m_slowdownMs > 0)
        return m_maxSpeed / maxSpeedFactor;

    return m_maxSpeed;
}

float Ship::getSteerPower() const
{
    return (m_steerPower * m_steerPowerMs / steerPowerMax);
}

void Ship::addAcceleration(int ms)
{
    m_accelerationMs = ms;
    m_slowdownMs = 0;
}

void Ship::addSlowdown(int ms)
{
    m_slowdownMs = ms;
    m_accelerationMs = 0;
}

bool Ship::isInSpace() const
{
    // Return true only if the ship is in space for some time
    if (m_inSpaceMs > inSpaceMax)
        return true;

    return false;
}

void Ship::setInSpace(bool inSpace)
{
    m_inSpace = inSpace;

    // If we are not in space, reset the time counter
    if (!inSpace)
        m_inSpaceMs = 0;
}

void Ship::setLocation(const Location &loc)
{
    if (!m_physicsLoaded)
        return;

    // We assume there is ground under the ship by default
    m_inSpace = false;
    m_inSpaceMs = 0;

    // Set new location using the physical engine
    btTransform transform = m_shipBody->getCenterOfMassTransform();
    transform.setOrigin(BtOgre::Convert::toBullet(loc.position));
    transform.setRotation(BtOgre::Convert::toBullet(loc.rotation));
    m_shipBody->setCenterOfMassTransform(transform);
    m_shipBody->setAngularVelocity(btVector3(0, 0, 0));
    m_shipBody->setLinearVelocity(btVector3(0, 0, 0));
}

void Ship::initPhysics()
{
    // Ship needs to be created first and also we can't init physics more than once
    if (!m_dataLoaded || m_physicsLoaded) {
        return;
    }

    // Create ship shape
    BtOgre::StaticMeshToShapeConverter converter(m_shipEntity);
    m_shipShape = converter.createBox();

    // Calculate inertia
    btVector3 inertia;
    m_shipShape->calculateLocalInertia(m_mass, inertia);

    // Create MotionState
    m_shipState = new BtOgre::RigidBodyState(m_shipNode);

    // Create ship body
    m_shipBody = new btRigidBody(m_mass, m_shipState, m_shipShape, inertia);
    m_shipBody->setFriction(0.0);
    m_shipBody->setRestitution(0.0);

    // Prevent body from sleeping
    m_shipBody->setActivationState(DISABLE_DEACTIVATION);

    // Now we need to get relative positions of bounding box corners (springs)
    if (m_shipShape->getShapeType() == BOX_SHAPE_PROXYTYPE)
    {
        const btBoxShape *box = static_cast<btBoxShape *>(m_shipShape);

        // Iterate over all 8 vertices
        for (int i = 0; i < box->getNumVertices(); ++i) {
            btVector3 vertex;
            box->getVertex(i, vertex);

            // We are interested only in the lower ones (under the ship)
            if (vertex.y() <= 0) {
                m_springs.push_back(vertex);
            }
        }
    }

    for (int i = 0; i < m_springs.size(); ++i) {
        Ogre::SceneNode *markerNodeLeft = m_sceneManager->createSceneNode();
        markerNodeLeft->setPosition(BtOgre::Convert::toOgre(m_springs[i]));

        Ogre::Entity *markerEntityLeft = m_sceneManager->createEntity("Marker.mesh");
        m_sceneManager->getRootSceneNode()->addChild(markerNodeLeft);
        markerNodeLeft->attachObject(markerEntityLeft);

        m_markers.push_back(markerNodeLeft);
    }

    m_physicsLoaded = true;
}
