/**
 *  @author Marian Ondrejicek (xondre02)
 *  @date   May 2012
 *
 *  Class that represents ship in the game.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef SHIP_H
#define SHIP_H

#include "levelobjects.h"

/** Ship in the game. Not only visuals, but also physical an other parameters.
  * Also provides methods for steering the ship and much more... */
class Ship
{
public:
    Ship();
    ~Ship();

    void create(const std::string &ship, const Location &loc);
    void setSceneManager(Ogre::SceneManager *manager) { m_sceneManager = manager; }

    /** Attach camera to ship's node, so camera follows when moving.
      * Never call this before calling created()! */
    void attachCamera(Ogre::Camera *cam);

    /** Update speed and direction based on time since last frame. */
    void update(double timeSinceLastFrame, bool paused);

    /** Set if we are accelerating or not. */
    void accelerate(bool val) { m_accelerate = val; }

    /** Return if we are accelerating or not. */
    bool isAccelerating() { return m_accelerate; }

    /** Set if we are slowing down or not. */
    void slowDown(bool val) { m_slowDown = val; }

    /** Return if we are slowing down or not. */
    bool isSlowingDown() { return m_slowDown; }

    /** Set if we are turning to left or not. */
    void steerLeft(bool val);

    /** Return if we are turning to left or not. */
    bool isSteeringLeft() { return m_steerLeft; }

    /** Set if we are turning to right or not. */
    void steerRight(bool val);

    /** Return if we are turning to right or not. */
    bool isSteeringRight() { return m_steerRight; }

    /** Return mass of the ship. */
    float getMass() const { return m_mass; }

    /** Return acceleration parameter. */
    float getAcceleration() const;
    
    /** Return maximal allowed speed of the ship. */
    float getMaxSpeed() const;

    /** Return steering power. */
    float getSteerPower() const;

    /** Return power of the breaks. */
    float getBreakPower() const { return m_breakPower; }

    /** Return current speed. */
    float getSpeed() const { return m_shipBody->getLinearVelocity().length(); }

    /** Ship has run over an accelerator. Extra boost for few milliseconds. */
    void addAcceleration(int ms);

    /** Ship has run over a slowdown. Extra slowdown for few milliseconds. */
    void addSlowdown(int ms);

    /** Is the ship in space or is it on the track? Returns true only
      * if the ship is in space for some time, not just few frames. */
    bool isInSpace() const;

    /** Set whether the ship is in space. This should be done in Physics. */
    void setInSpace(bool inSpace);

    /** Get the list of springs. */
    const btAlignedObjectArray<btVector3> &getSprings() const { return m_springs; }

    btAlignedObjectArray<Ogre::SceneNode*> &getMarkers() { return m_markers; }

    /** Get pointer to ship's body object. */
    btRigidBody *getBody() { return m_shipBody; }

    /** Return scene node of the ship. */
    Ogre::SceneNode *getNode() { return m_shipNode; }

    /** Place the ship into the space. */
    void setLocation(const Location &loc);

    /** Setup physics parameters. */
    void initPhysics();

private:
    Ogre::SceneManager *m_sceneManager; /**< Manages all objects in the scene. */
    Ogre::SceneNode *m_shipNode;
    Ogre::Entity *m_shipEntity;
    Ogre::ParticleSystem *m_partSystem;

    bool m_dataLoaded;              /**< Whether the data and visuals were successfully loaded. */
    bool m_physicsLoaded;           /**< Whether the ship's physics was successfully initialized. */

    bool m_accelerate;              /**< Should we accelerate in next update? */
    bool m_slowDown;                /**< Should we slow down in next update? */
    bool m_steerLeft;               /**< Should we go left in next update? */
    bool m_steerRight;              /**< Should we go right in next update? */
    
    float m_mass;                   /**< Weight of the ship. */
    float m_acceleration;           /**< Acceleration parameter. */
    float m_maxSpeed;               /**< Maximal allowed speed. */
    float m_steerPower;             /**< Steering power. */
    float m_breakPower;             /**< Power of the breaks. */

    int m_steerPowerMs;             /**< How long (ms) has the ship been steering. */
    int m_accelerationMs;           /**< Milliseconds of extra boost left. */
    int m_slowdownMs;               /**< Milliseconds of extra slowdown left. */

    bool m_inSpace;                 /**< Is the ship in space or on the track? */
    int m_inSpaceMs;                /**< How long is the ship in space (ms). */

    btRigidBody *m_shipBody;
    btCollisionShape *m_shipShape;
    BtOgre::RigidBodyState *m_shipState;

    /** Relative positions of four springs under the ship. */
    btAlignedObjectArray<btVector3> m_springs;
    btAlignedObjectArray<Ogre::SceneNode*> m_markers;
};

#endif  // SHIP_H
