/**
 *  @author Marian Ondrejicek (xondre02)
 *  @date   May 2012
 *
 *  Manages and plays all sounds and music.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "soundmanager.h"

// At the beginning, we have no instance
SoundManager *SoundManager::m_instance = NULL;

// How many songs are available in the game
const int gameSongsCount = 3;

SoundManager::SoundManager()
{
    m_muted = false;
    m_manager = OgreOggSound::OgreOggSoundManager::getSingletonPtr();
    m_manager->init();

    srand((unsigned)time(0));
}

SoundManager::~SoundManager()
{
    detach();
}

SoundManager *SoundManager::getSingleton()
{
    if (m_instance == NULL)
        m_instance = new SoundManager;

    return m_instance;
}

void SoundManager::attach(Ogre::SceneManager *sceneManager)
{
     m_manager->setSceneManager(sceneManager);
}

void SoundManager::detach()
{
    m_manager->setSceneManager(NULL);
    OgreOggSound::OgreOggSoundManager::getSingletonPtr()->destroyAllSounds();
}

void SoundManager::createMenuSounds()
{
    m_manager->createSound("MenuSong", "evro - Finding the way back.ogg", true, true, true);
}

void SoundManager::createGameSounds()
{
    m_manager->createSound("GameSong1", "evro - Space eel.ogg", true, false, true);
    m_manager->createSound("GameSong2", "evro - Stego's Back.ogg", true, false, true);
    m_manager->createSound("GameSong3", "evro - The Borg.ogg", true, false, true);
}

void SoundManager::setMute(bool muted)
{
    m_muted = muted;

    // Apply immediately
    if (m_muted)
        m_manager->pauseAllSounds();
    else
        m_manager->resumeAllPausedSounds();
}

bool SoundManager::isSongStopped() const
{
    OgreOggSound::OgreOggISound *sound = m_manager->getSound(m_song);

    if (sound)
        return sound->isStopped();

    return true;
}

void SoundManager::play(const std::string &name)
{
    m_manager->getSound(name)->play();

    if (m_muted)
        m_manager->pauseAllSounds();
}

void SoundManager::playRandom()
{
    // ID of last played song
    static int lastSong = 0;

    // Select one of three songs randomly, but don't play the last song again right now
    int currentSong;
    do {
        currentSong = rand() % gameSongsCount + 1;
    } while (currentSong == lastSong);
    lastSong = currentSong;

    // Select song name according to generated number
    if (currentSong == 1)
        m_song = "GameSong1";
    else if (currentSong == 2)
        m_song = "GameSong2";
    else
        m_song = "GameSong3";

    // Play the song
    m_manager->getSound(m_song)->play();

    if (m_muted)
        m_manager->pauseAllSounds();
}
